"""dotnet module commands"""

from devops_toolset.core.ValueDictsBase import ValueDictsBase
from devops_toolset.core.app import App
from enum import Enum

app: App = App()


class Log(Enum):
    """Log level for Azure CLI commands"""
    OFF = 0
    VERBOSE = 1
    DEBUG = 2


class Commands(ValueDictsBase):
    """ Commands for the project_types/azure module."""

    # Add your dotnet commands dictionaries here
    _commands = {
        "azure_cli_db_mysql_flexible_server_execute":
            "az mysql flexible-server execute -n {server_name} -u {admin_user} -p {admin_password} "
            "-d {database_name} {file_path} {query} {log}",
        "azure_cli_db_mysql_flexible_server_firewall_rule_create":
            "az mysql flexible-server firewall-rule create -n {server_name} -g {resource_group} "
            "-r {rule_name} --start-ip-address {start_ip_address} --end-ip-address {end_ip_address} {log}",
        "azure_cli_db_mysql_flexible_server_firewall_rule_delete":
            "az mysql flexible-server firewall-rule delete -n {server_name} -g {resource_group} "
            "-r {rule_name} -y {log}",
        "azure_cli_extension_list": "az extension list",
        "azure_cli_extension_add": "az extension add --name {name}",
        "azure_cli_login_service_principal": "az login --service-principal -u {user}  -p {secret} --tenant {tenant}",
        "azure_cli_logout": "az logout",
    }



"""Loads settings from the settings.json file"""

import pathlib
import os
import json
import pkg_resources


class Settings(object):
    """Application settings"""

    _DEVOPS: str = "devops_platforms"
    _LOCALES: str = "locales"
    _CORE: str = "core"
    _CONFIG_SETTINGS_FILE_NAME: str = "logging-config.json"
    _SETTINGS_FILE_NAME: str = "settings.json"
    _CURRENT_PATH: str = os.path.dirname(os.path.realpath(__file__))

    # defaults
    root_path: pathlib.Path = pathlib.Path(_CURRENT_PATH).parent.absolute()
    project_xml_path = root_path.parent.parent.absolute()
    devops_path: pathlib.Path = pathlib.Path.joinpath(root_path, _DEVOPS).absolute()
    locales_path: pathlib.Path = pathlib.Path.joinpath(root_path, _LOCALES).absolute()
    log_config_file_path: pathlib.Path = pathlib.Path(pkg_resources.resource_filename
                                                      (__name__, _CONFIG_SETTINGS_FILE_NAME))
    settings_path: pathlib.Path = pathlib.Path(pkg_resources.resource_filename(__name__, _SETTINGS_FILE_NAME))
    language: str = "en"
    platform: str = "azuredevops"
    platform_specific_path: pathlib.Path = pathlib.Path.joinpath(devops_path, platform).absolute()

    def __init__(self):
        """Loads settings"""

        self.load(self.settings_path.as_posix())
        self.platform_specific_path = pathlib.Path.joinpath(self.devops_path, self.platform).absolute()

    @staticmethod
    def read_settings_from_file(path: str):
        """Loads settings from settings.json

        Args:
            path: Path to read settings.json from
        """

        with open(path, "r") as settings_file:
            settings = json.load(settings_file)

        return settings

    def load(self, path: str):
        """Assigns settings"""

        settings = self.read_settings_from_file(path)

        # Add your setting mappings here
        self.language = settings["language"]
        self.platform = settings["platform"]



"""Azure DevOps REST API functionality"""

from devops_toolset.core.app import App
from devops_toolset.core.CommandsCore import CommandsCore
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.devops_platforms.azuredevops.Literals import Literals as PlatformSpecificLiterals
from devops_toolset.devops_platforms.azuredevops.commands import Commands as PlatformSpecificCommands
import devops_toolset.filesystem.paths
from devops_toolset.tools.xcoding64 import encode
from typing import Union
import logging
import requests


app: App = App()
literals = LiteralsCore([PlatformSpecificLiterals])
commands = CommandsCore([PlatformSpecificCommands])


def generate_authentication_header(user_name: str, access_token: str) -> dict:
    """Generates an authentication header.

    Args:
        user_name: Token name as defined in Azure Devops personal access tokens
        access_token: Token as defined in Azure DevOps personal access tokens.

    Returns:
        Authentication header
    """
    token_base64 = encode(f"{user_name}:{access_token}")
    basic_auth_token = f"Basic {token_base64}"
    return {"Authorization": basic_auth_token}


def call_api(command: str, user_name: str, access_token: str) -> requests.Response:
    """Calls the REST API.

    Args:
        command: Command to be executed.
        user_name: Token name as defined in Azure Devops personal access tokens
        access_token: Token as defined in Azure DevOps personal access tokens.

    Returns:
        Response
    """
    headers = generate_authentication_header(user_name, access_token)

    logging.info(literals.get("azdevops_command").format(command=command))

    response = requests.get(command, headers=headers)
    if response.status_code != 200:
        raise ValueError(literals.get("azdevops_status_code").format(status_code=response.status_code))
    return response


def get_build_list(organization: str, project: str, user_name: str, access_token: str) -> dict:
    """Gets a list of the builds for a project.

    Args:
        organization: Azure DevOps organization. i.e: aheadlabs
        project: Azure DevOps project. i.e: devops-toolset
        user_name: Token name as defined in Azure Devops personal access tokens
        access_token: Token as defined in Azure DevOps personal access tokens

    Returns:
        Build list based on
            https://docs.microsoft.com/en-us/rest/api/azure/devops/build/builds/list#build
    """
    command = commands.get("azdevops_rest_get_build_list").format(organization=organization, project=project)
    return call_api(command, user_name, access_token).json()


def get_last_build_id(organization: str, project: str, user_name: str, access_token: str) -> Union[int, None]:
    """Returns the id of the last build.

    Args:
        organization: Azure DevOps organization. i.e: aheadlabs
        project: Azure DevOps project. i.e: devops-toolset
        user_name: Token name as defined in Azure Devops personal access tokens
        access_token: Token as defined in Azure DevOps personal access tokens

    Returns:
        Id of the build
    """
    build_list = get_build_list(organization, project, user_name, access_token)

    if build_list["count"] > 0:
        return build_list["value"][0]["id"]
    return None


def get_build(organization: str, project: str, build_id: int, artifact_name: str,
              user_name: str, access_token: str) -> dict:
    """Gets a build for a project.

    Args:
        organization: Azure DevOps organization. i.e: aheadlabs
        project: Azure DevOps project. i.e: devops-toolset
        build_id: Build id to obtained
        artifact_name: Name of the artifact to be obtained
        user_name: Token name as defined in Azure Devops personal access tokens
        access_token: Token as defined in Azure DevOps personal access tokens
    """
    command = commands.get("azdevops_rest_get_build").format(
        organization=organization, project=project, build_id=build_id, artifact_name=artifact_name)
    return call_api(command, user_name, access_token).json()


def get_last_build(organization: str, project: str, artifact_name: str,
                   user_name: str, access_token: str) -> dict:
    """Gets a build for a project.

    Args:
        organization: Azure DevOps organization. i.e: aheadlabs
        project: Azure DevOps project. i.e: devops-toolset
        artifact_name: Name of the artifact to be obtained
        user_name: Token name as defined in Azure Devops personal access tokens
        access_token: Token as defined in Azure DevOps personal access tokens
    """
    build_id = get_last_build_id(organization, project, user_name, access_token)
    return get_build(organization, project, build_id, artifact_name, user_name, access_token)


def get_artifact(organization: str, project: str, build_id: int, artifact_name: str, destination_path: str,
                 user_name: str, access_token: str):
    """Gets a build for a project.

    Args:
        organization: Azure DevOps organization. i.e: aheadlabs
        project: Azure DevOps project. i.e: devops-toolset
        build_id: Build id to obtained
        artifact_name: Name of the artifact to be obtained
        destination_path: Path where the artifact will be saved.
        user_name: Token name as defined in Azure Devops personal access tokens
        access_token: Token as defined in Azure DevOps personal access tokens
    """
    build = get_build(organization, project, build_id, artifact_name, user_name, access_token)
    download_url = build["resource"]["downloadUrl"]
    if download_url:
        headers = generate_authentication_header(user_name, access_token)
        devops_toolset.filesystem.paths.download_file(
            build["resource"]["downloadUrl"], destination_path, f"{artifact_name}.zip", headers)


def get_last_artifact(organization: str, project: str, artifact_name: str, destination_path: str,
                      user_name: str, access_token: str):
    """Gets the last build for a project.

    Args:
        organization: Azure DevOps organization. i.e: aheadlabs
        project: Azure DevOps project. i.e: devops-toolset
        artifact_name: Name of the artifact to be obtained
        destination_path: Path where the artifact will be saved.
        user_name: Token name as defined in Azure Devops personal access tokens
        access_token: Token as defined in Azure DevOps personal access tokens
    """
    last_build_id = get_last_build_id(organization, project, user_name, access_token)
    if last_build_id:
        get_artifact(organization, project, last_build_id, artifact_name, destination_path, user_name, access_token)


if __name__ == "__main__":
    help(__name__)



"""wordpress module commands"""

from devops_toolset.core.ValueDictsBase import ValueDictsBase
from devops_toolset.core.app import App

app: App = App()


class Commands(ValueDictsBase):
    """Commands for the wordpress module."""

    # Add your wordpress literal dictionaries here
    _commands = {
        "wpcli_config_create": "wp config create --path={path} --dbhost={db_host} --dbname={db_name} "
                               "--dbuser={db_user} --dbpass={db_pass} --dbprefix={db_prefix} --dbcharset={db_charset} "
                               "--dbcollate={db_collate} --force {skip_check} {debug_info}",
        "wpcli_config_set": "wp config set {name} {value} {raw} --type={type} --path={path} --separator=\\n "
                            "{debug_info}",
        "wpcli_core_download": "wp core download --version={version} --locale={locale} --path={path} "
                               "{skip_content} {debug_info}",
        "wpcli_core_install": "wp core install --path={path} --url={url} --title=\"{title}\" --admin_user={admin_user} "
                              "--admin_email={admin_email} {admin_password} {skip_email} {debug_info}",
        "wpcli_core_version": "wp core version --path={path}",
        "wpcli_db_create": "wp db create {db_user} {db_pass} --path={path} {debug_info}",
        "wpcli_db_export": "wp db export \"{core_dump_path}\" --path={path} --extended-insert=false {debug_info}",
        "wpcli_db_reset": "wp db reset --path={path} {yes} {debug_info}",
        "wpcli_db_import": "wp db import {file} --path={path} {debug_info}",
        "wpcli_db_delete_transient": "wp transient delete --all --path={path}",
        "wpcli_db_query_create_user":
            "wp db query \"create user '{user}'@'{host}' identified by '{password}'\" "
            "--dbuser={admin_user} --dbpass={admin_password} --path={path}",
        "wpcli_db_query_db_exists": "wp db query \"select exists (select 1 from information_schema.schemata where "
                                    "schema_name = '{schema}')\" --dbuser={admin_user} --dbpass={admin_password} "
                                      "--path={path}",
        "wpcli_db_query_grant": "wp db query \"grant {privileges} on {schema}.* to '{user}'@'{host}'\" "
            "--dbuser={admin_user} --dbpass={admin_password} --path={path}",
        "wpcli_db_query_user_exists": "wp db query \"select exists (select 1 from mysql.user where user='{user}')\" "
                                      "--dbuser={admin_user} --dbpass={admin_password} "
                                      "--path={path}",
        "wpcli_post_list_ids": "wp post list --post_type={post_type} --path={path} --format=ids",
        "wpcli_post_delete_post_type": "wp post delete {id_list} --force --path={path} {debug_info}",
        "wpcli_eval": "wp eval \"{php_code}\" --path={path}",
        "wpcli_export": "wp export --path=\"{path}\" --dir=\"{destination_path}\" "
                        "--filename_format={date}_UTC-content{suffix}.xml",
        "wpcli_import": "wp import \"{file}\" --authors={authors} --path=\"{path}\" {debug_info}",
        "wpcli_info": "wp --info",
        "wpcli_option_add": "wp option add {option_name} \"{option_value}\" {autoload} --path={path} {debug_info}",
        "wpcli_option_get": "wp option get {option_name} --path={path} {debug_info}",
        "wpcli_option_update": "wp option update {option_name} \"{option_value}\" {autoload} "
                               "--path={path} {debug_info}",
        "wpcli_plugin_install": "wp plugin install {source} --path={path} {force} {activate} {debug_info}",
        "wpcli_rewrite_structure": "wp rewrite structure {structure} --path={path} {debug_info}",
        "wpcli_theme_install": "wp theme install {source} --path={path} {activate} --force {debug_info}",
        "wp_theme_src_build": "gulp build --theme-slug=\"{theme_slug}\" --dist=\"{path}\"",
        "wp_theme_src_watch":
            "gulp watch --theme-slug=\"{theme_slug}\" --dev-proxy=\"{local_web_server}\" --wordpress-path=\"{path}\"",
        "wp_user_create": "wp user create {user_login} {user_email} "
                          "{role} {display_name} {first_name} {last_name} {send_email} --path={path} {debug_info}",
        "wp_user_get": "wp user get {user_login} --format=json --path={path} {debug_info}"
    }



"""Logging Formatter to add colors to log records using Colorama """
import logging
from colorama import Fore, Back, Style


class ColorFormatter(logging.Formatter):
    """Logging Formatter to add colors to log records """

    def __init__(self, format_str):
        """ Initializes the color formatter with color data and takes the format string as parameter"""
        grey = Fore.WHITE
        yellow = Fore.YELLOW
        red = Fore.RED
        reset = Style.RESET_ALL

        self.FORMATS = {
            logging.DEBUG: grey + format_str + reset,
            logging.INFO: grey + format_str + reset,
            logging.WARNING: yellow + format_str + reset,
            logging.ERROR: red + format_str + reset,
            logging.CRITICAL: Fore.BLACK + Back.RED + format_str + reset
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)



""" Creates an opinionated basic structure for a WordPress project. """

import devops_toolset.filesystem.paths as path_tools
import logging
import os
import pathlib
import requests
from devops_toolset.core.app import App
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.project_types.wordpress.Literals import Literals as WordpressLiterals
from typing import Union

app: App = App()
literals = LiteralsCore([WordpressLiterals])


class BasicStructureStarter(object):
    """ Set of methods used to create a basic structure for a Wordpress project """

    def add_item(self, item, base_path_str: str) -> None:
        """ Creates the item (file or dir) in the current filesystem

         Args:
                item: the object containing the type and the value of the content.
                base_path_str: the path which from the structure will be built

        """

        # Set paths
        base_path = pathlib.Path(base_path_str)
        final_path = pathlib.Path.joinpath(base_path, item["name"])
        has_condition = "condition" in item
        if has_condition:
            child_condition = self.condition_met(item, base_path_str)
        else:
            child_condition = True

        # Only if the item DOES NOT exist and condition is met
        if not path_tools.is_valid_path(str(final_path), True) and child_condition and "type" in item:
            # Create item
            if item["type"] == "directory":
                os.mkdir(final_path)
                logging.debug(literals.get("wp_directory_created").format(directory=final_path))
            elif item["type"] == "file":
                with open(final_path, "w", newline="\n") as new_file:
                    if "default_content" in item:
                        default_content = self.get_default_content(item["default_content"], False)
                        new_file.write(default_content)
                    logging.debug(literals.get("wp_file_created").format(file=final_path))
            elif item["type"] == "bfile" and "default_content" in item:
                default_content = self.get_default_content(item["default_content"], True)
                with open(final_path, "wb") as new_file:
                    new_file.write(default_content)
                    logging.debug(literals.get("wp_file_created").format(file=final_path))

        # Iterate through children if any
        if "children" in item:
            for child in item["children"]:
                self.add_item(child, final_path)

    @staticmethod
    def condition_met(item, base_path: str) -> bool:
        """ Returns the result (True of False) of the condition contained on the item

        Args:
            item: the item inspected for conditions
            base_path: parent dir to check
        """

        if "condition" in item and item["condition"] == "when-parent-not-empty":
            return path_tools.is_empty_dir(str(pathlib.Path(base_path)))
        # Default behaviour
        return True

    @staticmethod
    def get_default_content(item, is_binary: bool = False) -> Union[str, bytes]:
        """ Gets the default content of the files based on the json item passed

            Args:
                item : the object containing the type and the value of the default content
                is_binary: if True returns content as bytes, otherwise as str

            Returns:
                Content in text or binary format
        """
        if item["source"] == "raw":
            logging.debug(literals.get("wp_write_default_content").format(
                file="",
                source=f"raw data => {item['source']}"
            ))
            return item["value"]
        elif item["source"] == "from_file":
            with open(item["value"], "r") as default_content_file:
                logging.debug(literals.get("wp_write_default_content").format(
                    file="",
                    source=f"file => {item['source']}"
                ))
                return default_content_file.read()
        elif item["source"] == "from_url":
            logging.debug(literals.get("wp_write_default_content").format(
                file="",
                source=f"URL => {item['source']}"
            ))
            response = requests.get(item["value"])
            return response.content if is_binary else response.text



"""Provides tools for managing the AWS CloudFront service"""

from devops_toolset.core.app import App
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.project_types.aws.Literals import Literals as AwsLiterals

import boto3
import logging
import time

app: App = App()
cloudfront = boto3.client("cloudfront")
literals = LiteralsCore([AwsLiterals])


def create_invalidation(distribution_id: str, invalidation_paths: list[str] = ["/*"]):
    """Creates a CloudFront invalidation.

    Args:
        distribution_id: Id of the CloudFront distrbution where the
            invalidation will be created.
        invalidation_paths: List of all the paths to be invalidated.
    """

    response = cloudfront.create_invalidation(
        DistributionId=distribution_id,
        InvalidationBatch={
            "Paths": {
                "Quantity": len(invalidation_paths),
                "Items": invalidation_paths
            },
            "CallerReference": str(time.time()).replace(".", "")
        }
    )

    logging.info(literals.get("cloudfront_invalidation_created").format(id=response['Invalidation']['Id']))


if __name__ == "__main__":
    help(__name__)



"""wordpress module literals"""

from devops_toolset.core.ValueDictsBase import ValueDictsBase
from devops_toolset.core.app import App

app: App = App()


class Literals(ValueDictsBase):
    """ValueDicts for the wordpress module."""

    _titles = {
        "wp_title_generate_wordpress": _("Generate WordPress site"),
        "wp_title_bootstrap_repository": _("Bootstrap WordPress repository"),
        "wp_title_wordpress_files": _("WordPress core"),
        "wp_title_wordpress_new_repo": _("WordPress\nexisting repo"),
        "wp_title_wordpress_rollback_db": _("WordPress\ndatabase rollback"),
    }
    _info = {
        "wp_created_project_structure": _("Finished creation of the project structure."),
        "wp_creating_project_structure": _("Starting to create the project structure."),
        "wp_created_theme_structure": _("Finished creation of the development theme {theme_name} structure."),
        "wp_project_structure_creating_from_file": ("Creating project structure from {file_name}"),
        "wp_project_structure_creating_from_default_file": _("Creating default project structure from "
                                                           "{resource}"),
        "wp_theme_structure_creating_from_file": _("Creating development theme [{theme_name}] structure from "
                                                   "{file_name}"),
        "wp_theme_structure_creating_from_default_file": _("Creating default development theme structure from "
                                                   "{resource}"),
        "wp_default_files": _("These are the default files from the GitHub repository:"),
        "wp_directory_created": _("Directory created: {directory}"),
        "wp_environment_file_used": _("The following environment file will be used: {file}"),
        "wp_file_created": _("File created: {file}"),
        "wp_file_deleted": _("File deleted: {file}"),
        "wp_got_db_admin_user": _("I got the following database admin user from the environment configuration file: "
                                  "{db_admin_user}"),
        "wp_init_git_repo": _("Do you want me to initialize a local Git repository for you?"),
        "wp_looking_for_src_themes": _("Looking for development themes to build..."),
        "wp_no_src_themes": _("No development themes to build. Skipping this step."),
        "wp_parsing_theme_metadata": _("Parsing theme metadata..."),
        "wp_parsing_theme_regex": _("Regex to match theme metadata content is: {regex}"),
        "wp_parsing_theme_no_matches_found": _("No match/es found for token: {token}"),
        "wp_parsing_theme_matches_found": _("Matched content {content} for token {token}"),
        "wp_required_file_paths_not_found": _("No required file paths were found"),
        "wp_required_file_paths_found": _("The following file paths were found:"),
        "wp_root_path": _("The root path is: {path}"),
        "wp_themes_install_manually": _("Please, install the theme/s manually."),
        "wp_plugin_path": _("The plugin path is: {path}"),
        "wp_theme_path": _("The theme path is: {path}"),
        "wp_themes_path": _("The themes path is: {path}"),
        "wp_use_default_files": _("Do you want me to use the default ones instead for those missing files?"),
        "wp_wordpress_path": _("The WordPress path is: {path}"),
        "wp_wpcli_add_ev": _("Remember to add the path to the PATH or BIN, depending on your operating system."),
        "wp_wpcli_backup_create_before": _("Creating database backup..."),
        "wp_wpcli_config_created_ok": _("File wp-config.php created successfully."),
        "wp_wpcli_core_install_before": _("Preparing wordpress core files to install..."),
        "wp_wpcli_creating_config": _("Creating wp-config.php..."),
        "wp_wpcli_db_create_before":
            _("Creating Wordpress database (by default, user_name and host will be taken from wp_config)"),
        "wp_wpcli_db_import_before": _("Importing database dump from file:"),
        "wp_wpcli_db_export_before": _("Exporting database dump to: {path}"),
        "wp_wpcli_db_query_user_creating": _("Creating database user {user} for host {host}"),
        "wp_wpcli_db_query_user_granting": _("Granting database user {user} for host {host} the following privileges "
                                             "on schema {schema}: {privileges}"),
        "wp_wpcli_db_reset_before": _("Resetting the database (drop and create)..."),
        "wp_wpcli_delete_transients": _("Transients are going to be deleted"),
        "wp_wpcli_downloading": _("Downloading WP-CLI from {url}"),
        "wp_wpcli_downloading_content": _("Downloading WordPress content (default themes and plugins): {content}"),
        "wp_wpcli_downloading_path": _("WordPress core destination path: {path}"),
        "wp_wpcli_downloading_wordpress": _("Downloading WordPress core files version {version}, locale {locale}"),
        "wp_wpcli_downloading_wordpress_ok": _("WordPress core files were successfully downloaded."),
        "wp_wpcli_export": _("Exporting WordPress content to {path}"),
        "wp_wpcli_import_after": _("{type} content imported successfully. "),
        "wp_wpcli_import_before": _("Importing {type} content..."),
        "wp_wpcli_import_error": _("There was an error when importing {type} content..."),
        "wp_wpcli_info": _("Here is the WP-CLI information:"),
        "wp_wpcli_install_ok": _("WP-CLI installation was successful."),
        "wp_wpcli_option_add_before": _("Adding database option {option_name}..."),
        "wp_wpcli_option_update_before": _("Updating database option {option_name}..."),
        "wp_wpcli_option_skipping": _("Skipping option update for {option_name} since the new value is the same as the "
                                      "existing one."),
        "wp_wpcli_plugin_install_before": _("Installing plugin {plugin_name}..."),
        "wp_wpcli_plugin_install_error": _("An error occurred installing plugin {plugin_name}..."),
        "wp_wpcli_post_delete_post_type_before": _("Deleting posts of type {post_type}..."),
        "wp_wpcli_setting_value_ok": _("Config value {name} set as {value}"),
        "wp_wpcli_theme_install_before": _("Installing wordpress theme {theme_name}"),
        "wp_wpcli_user_creating": _("Creating WordPress user {user}..."),
        "wp_wpcli_user_created": _("Created WordPress user {user}."),
        "wp_write_default_content": _("Writing default content to file \"{file}\" from {source}"),
        "wp_gulp_build_before": _("Gulp build task has launched for theme {theme_slug}."),
        "wp_gulp_build_after": _("Gulp build task has completed successfully for theme {theme_slug}."),
    }
    _warnings = {
        "mysql_user_exists_grant_privileges_manually": _("User {user} exists in the {host} host and will not be "
                                                         "altered. Please grant the following privileges manually to "
                                                         "{host} for this user in the {schema} schema: "
                                                         "(db privileges) {db_privileges}; "
                                                         "(global privileges) {global_privileges};"),
        "mysql_db_exists_skipping_creation": _("Database {schema} exists. I will not create any database..."),
        "wp_wpcli_export_db_skipping_as_set": _("I am skipping the {dump} database dump as configured in settings..."),
        "wp_wpcli_user_exists": _("User {user} already exists. Skipping user creation..."),
    }
    _errors = {
        "wp_checking_devops_toolset": _("Checking for devops-toolset in: {path}"),
        "wp_config_file_bad_themes_number": _("There are {number} themes in the configuration file (must be 1 or 2)."),
        "wp_config_file_only_one_activated_theme": _("You can only activate one theme and there are {number} themes "
                                                     "marked to be activated. Please, check the configuration file."),
        "wp_current_version": _("Current version: {version}"),
        "wp_devops_toolset_needs_update": _("devops-toolset needs to be updated."),
        "wp_devops_toolset_not_found": _("devops-toolset not found in required path: {path} "),
        "wp_devops_toolset_obtained": _("devops-toolset has been successfully downloaded on: {path}"),
        "wp_devops_toolset_obtaining": _("Obtaining latest version of devops-toolset from: {resource}"),
        "wp_devops_toolset_up_to_date": _("devops-toolset is up to date."),
        "wp_env_gt1": _("There are more than 1 matching environments in the environments JSON file."),
        "wp_env_not_found": _("Environment not found in the environments list."),
        "wp_env_x_not_found": _("Environment {environment} not found in the environments list."),
        "wp_environment_name_mandatory": _("environment-name is a mandatory parameter. I cannot continue."),
        "wp_environment_name_not_found": _("Environment name was not found. I cannot continue."),
        "wp_environment_path_mandatory": _("environment-path is a mandatory parameter. I cannot continue."),
        "wp_environment_path_not_found": _("Environment file path was not found. I cannot continue."),
        "wp_environment_x_found_multiple": _("Multiple environments found with name {environment}. "
                                             "I take the first one."),
        "wp_file_not_found": _("The following file was not found: {file}"),
        "wp_latest_version": _("Latest version: {version}"),
        "wp_not_dir": _("Path must be a dir, not a file."),
        "wp_non_valid_dir_path": _("Path must be an existent dir."),
        "wp_required_files_mandatory": _("Required files are mandatory. I cannot continue."),
        "wp_required_files_not_found": _("The required files where not found at {path}."),
        "wp_required_files_not_found_detail": _("The following required files where not found at {path}:"),
        "wp_theme_path_not_exist": _("The following theme path does not exist: {path}"),
        "wp_theme_feed_no_info": _("The {theme} theme has a source type of feed, but it has no feed configuration. "
                                   "Please, check the configuration file."),
        "wp_wordpress_path_mandatory": _("wordpress-path is a mandatory parameter. I cannot continue."),
        "wp_wpcli_config_create_err": _("File wp-config.php cannot be created."),
        "wp_wpcli_config_set_value_err": _("Cannot set {name} property as {value}"),
        "wp_wpcli_core_install_error": _("Wordpress installation could not be done."),
        "wp_wpcli_core_version_already_downloaded": _("Wordpress {version} is already present in the specified path. "
                                                      "Skipping download..."),
        "wp_wpcli_db_export_error": _("Database dump could not be exported due to an error."),
        "wp_wpcli_db_import_error": _("Dump file could not be imported due to an error."),
        "wp_wpcli_db_query_user_creating_err": _("Database user {user} for host {host} could not be created"),
        "wp_wpcli_db_query_user_granting_err": _("Privileges could not be granted to database user {user} for host "
                                                 "{host} on schema {schema}"),
        "wp_wpcli_db_reset_error": _("Database could not be reset due to an error."),
        "wp_wpcli_delete_transients_err": _("Transients could not be deleted"),
        "wp_wpcli_downloading_wordpress_err": _("WordPress core files could not be downloaded."),
        "wp_wpcli_export_err": _("WordPress content could not be exported to {path}"),
        "wp_wpcli_option_add_error":
            _("Database option {option_name} cannot be added due to an error."),
        "wp_wpcli_option_update_error":
            _("Database option {option_name} cannot be set to {option_value} due to an error."),
        "wp_wpcli_plugin_install_err": _("Plugin {plugin_name} could not be installed due to an error."),
        "wp_wpcli_post_delete_post_type_err": _("Unable to delete content from type {post_type} due to an error."),
        "wp_wpcli_user_creating_err": _("An error occurred creating the user {user}."),
        "wp_src_theme_not_found": _("Create development theme was called but no src themes found. Please check your "
                                    "themes configuration and try again."),
        "wp_wpcli_theme_install_error": _("Theme {theme_name} could not be installed."),
        "wp_gulp_build_error":
            _("Gulp build task has encountered an error for theme {theme_slug}. "
              "Please check above logs for more details."),
    }



"""Provides tools for managing the AWS S3 service"""

import boto3
import devops_toolset.filesystem.paths as paths
import logging
import os
import pathlib

from datetime import date
from devops_toolset.core.app import App
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.project_types.aws.Literals import Literals as AwsLiterals


app: App = App()
s3 = boto3.client("s3")
literals = LiteralsCore([AwsLiterals])


def get_objects_from_bucket(bucket_name: str, keys: list[str], destination_path: str):
    """Downloads objects that match the specified keys from the bucket.

    Args:
        bucket_name: S3 bucket name to get the objects from.
        keys: Keys of the objects to be downloaded.
        destination_path: Path where the objects will be downloaded to.
    """

    if not paths.is_valid_path(destination_path, True):
        raise ValueError

    if len(keys) == 0:
        raise ValueError

    destination_path_obj = pathlib.Path(destination_path)

    logging.info(literals.get("s3_downloading_objects_from_s3_bucket").format(
        number=len(keys),
        bucket=bucket_name,
        destination=str(destination_path_obj)
    ))

    for key in keys:
        object_destination_path_obj = pathlib.Path.joinpath(destination_path_obj, key)
        with open(object_destination_path_obj, "wb") as file:
            s3.download_fileobj(bucket_name, key, file)
            logging.info(literals.get("s3_downloaded_object_from_s3_bucket").format(
                name=key,
                bucket=bucket_name,
                destination=str(object_destination_path_obj)
            ))


def get_filtered_objects_from_bucket(bucket_name: str, object_prefix: str, destination_path: str):
    """Downloads filtered objects that match the specified keys and prefix from
    the bucket.

    Args:
        bucket_name: S3 bucket name to get the objects from.Hello world!
        object_prefix: Only the object keys that match this prefix will be
            retrieved.
        destination_path: Path where the objects will be downloaded to.
    """

    if not paths.is_valid_path(destination_path, True):
        raise ValueError()

    if object_prefix is None or object_prefix == "":
        raise ValueError()

    object_list = list_objects_in_bucket(bucket_name, object_prefix)
    key_list = list(map(lambda x: x["Key"], object_list))
    get_objects_from_bucket(bucket_name, key_list, destination_path)


def list_objects_in_bucket(bucket_name: str, object_prefix: str = "") -> list:
    """Gets all objects that match the criteria from a S3 bucket.

    Args:
        bucket_name: S3 bucket name to get the object list from.
        object_prefix: Only the object keys that match this prefix will be
            retrieved.

    Returns:
        Returns the list of all the objects in the bucket.
    """

    object_list: list = []

    paginator = s3.get_paginator("list_objects_v2")
    pages = paginator.paginate(Bucket=bucket_name, Prefix=object_prefix)
    for page in pages:
        object_list.extend(page["Contents"])

    logging.info(literals.get("s3_got_x_objects_from_s3_bucket").format(
        number=len(object_list),
        bucket=bucket_name
    ))

    return object_list


def put_object_to_bucket(bucket_name: str, local_path: str, destination_key: str):
    """Uploads an object to the S3 bucket.

    Args:
        bucket_name: S3 bucket name to get the object list from.
        local_path: Path to the file to be uploaded.
        destination_key: Path where the objects will be uploaded to.
    """

    if not paths.is_valid_path(local_path, True):
        raise ValueError()

    with open(local_path, "rb") as file:
        content = file.read()

    s3.put_object(
        Bucket=bucket_name,
        Body=content,
        Key=destination_key
    )

    logging.info(literals.get("s3_uploaded_object_to_s3_bucket").format(
        object_key=destination_key,
        bucket=bucket_name
    ))


def put_bulk_objects_to_bucket(bucket_name: str, local_path: str, glob: str, destination_key_prefix: str = ""):
    """Uploads multiple objects to the S3 bucket based on a glob expression.

    Args:
        bucket_name: S3 bucket name to upload objects to.
        local_path: Path to the base directories where files are located.
        glob: Glob expression used to select the files to be uploaded.
        destination_key_prefix: Prefix added to every object key.
    """

    if not paths.is_valid_path(local_path, True):
        raise ValueError()

    file_list: list = paths.get_file_paths_in_tree(local_path, glob)

    for file in file_list:
        object_key = f"{destination_key_prefix}{os.path.basename(file)}"
        put_object_to_bucket(bucket_name, str(file), object_key)

if __name__ == "__main__":
    help(__name__)



"""devops_toolset.devops_platforms.aws module literals."""

from devops_toolset.core.ValueDictsBase import ValueDictsBase
from devops_toolset.core.app import App

app: App = App()


class Literals(ValueDictsBase):
    """ValueDicts for the devops_toolset.devops_platforms.aws module."""

    # Add your core literal dictionaries here
    _info = {
        "aws_command": _("Command: {command}"),
        "platform_created_ev": _("Created environment variable {key} with value {value}"),
    }
    _errors = {}



"""Artifacts-related functionality"""

import devops_toolset.tools.cli as cli
from devops_toolset.core.app import App
from devops_toolset.core.CommandsCore import CommandsCore
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.devops_platforms.azuredevops.Literals import Literals as PlatformSpecificLiterals
from devops_toolset.devops_platforms.azuredevops.commands import Commands as PlatformSpecificCommands
import logging


app: App = App()
literals = LiteralsCore([PlatformSpecificLiterals])
commands = CommandsCore([PlatformSpecificCommands])


def download_artifact_from_feed(feed_config: dict, destination_path: str, **kwargs):
    """Downloads an artifact from a feed

    Args:
        feed_config: Feed configuration.
        destination_path: Path where the artifact will be downloaded.
        kwargs: Platform-specific arguments.
    """
    if "azdevops_token" not in kwargs:
        logging.error(literals.get("azdevops_token_not_found"))
        logging.warning(literals.get("azdevops_download_package_manually"))
        return

    azdevops_token = kwargs["azdevops_token"]
    cli.call_subprocess(commands.get("azdevops_cli_login")
                        .format(token=azdevops_token, organization=feed_config["organization_url"]))

    cli.call_subprocess(commands.get("azdevops_cli_universal_download")
                        .format(feed=feed_config["name"],
                                name=feed_config["package"],
                                path=f"\"{destination_path}\"",
                                version=feed_config["version"],
                                organization=feed_config["organization_url"]))


if __name__ == "__main__":
    help(__name__)



"""linux module literals"""

from devops_toolset.core.app import App
from devops_toolset.core.ValueDictsBase import ValueDictsBase

app: App = App()


class Literals(ValueDictsBase):
    """ValueDicts for the linux module."""

    _info = {
        "deb_package_install_pre": "Preparing to install: {package}. Please wait...",
        "deb_package_install_post": "Package {package} has been succesfully installed.",
        "edit_in_place_post": "Replaced {search_for}->{replace_with} completed in {path}"
    }
    _errors = {
        "deb_package_install_err": "An error occurred when trying to install {package}. Please check the logs.",
        "file_not_exist_err": "File {path} not found.",
        "edit_in_place_err": "Something went wrong replacing {search_for}->{replace_with} in {path}"
    }



"""Generates a WordPress Git repository for local development.

Git repositories should not contain downloadable or third-party files, but you need them for your site to work.
These are the type of files that won't be pushed to the repository and we will download/generate here:
  - WordPress core files
  - This toolset's
  - WordPress themes (parent themes)

Args:
    --project-path: Path to the WordPress installation.
    --environment-path: Path to the environment JSON file.
    --environment-name: Environment name.
    --db-user-password: Password for the database user.
    --db-admin-password: Password for the WordPress admin user.
"""

import argparse
import os
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.core.app import App
from devops_toolset.project_types.wordpress import generate_wordpress
from devops_toolset.project_types.wordpress.Literals import Literals as WordpressLiterals
from devops_toolset.tools import git

app: App = App()
literals = LiteralsCore([WordpressLiterals])


def main(project_path: str, db_user_password: str, db_admin_password: str, wp_admin_password: str,
         environment: str, additional_environments: list, additional_environment_db_user_passwords: list,
         create_db: bool, skip_partial_dumps: bool, skip_git: bool, **kwnargs):
    """Generates a WordPress Git repository for local development."""

    # Change the working directory
    os.chdir(project_path)

    # Initialize a local Git repository?
    git.git_init(project_path, skip_git)

    # Generate wordpress with the required data
    generate_wordpress.main(project_path, db_user_password, db_admin_password, wp_admin_password, environment,
                            additional_environments, additional_environment_db_user_passwords, create_db,
                            skip_partial_dumps, **kwnargs)

    # Move initial required files to .devops
    # TODO(ivan.sainz) Move initial required files to .devops

    # Commit git repository
    git.git_commit(skip_git)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("project-path", action=tools.argument_validators.PathValidator)
    parser.add_argument("--additional-environments", default="")
    parser.add_argument("--additional-environment-db-user-passwords", default="")
    parser.add_argument("--create-db", action="store_true", default=False)
    parser.add_argument("--db-user-password", required=True)
    parser.add_argument("--db-admin-password", required=True)
    parser.add_argument("--environment", default="localhost")
    parser.add_argument("--skip-git", action="store_true", default=False)
    parser.add_argument("--skip-partial-dumps", action="store_true", default=False)
    parser.add_argument("--wp-admin-password", required=True)
    kwargs = {}
    args, args_unknown = parser.parse_known_args()
    for kwarg in args_unknown:
        splitted = str(kwarg).split("=")
        kwargs[splitted[0]] = splitted[1]
    tools.cli.print_title(literals.get("wp_title_generate_wordpress"))
    main(args.project_path, args.db_user_password, args.db_admin_password, args.wp_admin_password,
         args.environment,
         args.additional_environments.split(","),
         args.additional_environment_db_user_passwords.split(","),
         args.create_db, args.skip_partial_dumps, args.skip_git, **kwargs)



"""AWS module literals"""

from devops_toolset.core.app import App
from devops_toolset.core.ValueDictsBase import ValueDictsBase

app: App = App()


class Literals(ValueDictsBase):
    """ValueDicts for the AWS module."""

    _info = {
        "cloudfront_invalidation_created": _("Created CloudFront invalidation with id {id}"),
        "s3_downloaded_object_from_s3_bucket":
            _("Downloaded object with key {name} from bucket {bucket} to {destination}"),
        "s3_downloading_objects_from_s3_bucket":
            _("Downloading {number} objects from bucket {bucket} to {destination}"),
        "s3_got_x_objects_from_s3_bucket": _("Got a list of {number} objects from the {bucket} bucket"),
        "s3_uploaded_object_to_s3_bucket": _("Uploaded object {object_key} to {bucket} bucket"),
    }
    _errors = {}



"""filesystem module literals"""

from devops_toolset.core.ValueDictsBase import ValueDictsBase
from devops_toolset.core.app import App

app: App = App()


class Literals(ValueDictsBase):
    """ValueDicts for the filesystem module."""

    _titles = {}
    _info = {
        "fs_composer_path_is": _("Composer file path is {path}."),
        "fs_file_content": _("File content => {content}."),
        "fs_file_moving": _("Moving file {origin_file_path} to {destination_file_path}."),
        "fs_file_path": _("File path is: {path}"),
        "fs_file_path_not_valid": _("File path is not valid."),
        "fs_file_path_does_not_exist": _("File path does not exist."),
        "fs_project_path_is": _("Project path is {path}."),
        "fs_strip_bom": _("I am stripping BOM character from file \"{path}\""),
        "fs_writing_file": _("Writing file \"{path}\""),
        "fs_zip_added_file": _("[{zip_file_name}] Added file: {added_file}."),
    }
    _errors = {
        "fs_not_dir": _("Path must be a dir, not a file."),
        "list_length_zero": _("List length is 0."),
        "list_length_higher_than": _("List length is higher than {length} and must be lower."),
    }



"""dotnet module commands"""

from devops_toolset.core.ValueDictsBase import ValueDictsBase
from devops_toolset.core.app import App

app: App = App()


class Commands(ValueDictsBase):
    """ Commands for the dotnet module."""

    # Add your dotnet commands dictionaries here
    _commands = {
        "dotnet_build":
            "dotnet build {force} {with_restore} --configuration={configuration} --framework={framework} "
            "--runtime={runtime} {debug} {output} {path}",
        "dotnet_ef_database_drop":
            "dotnet ef database drop --force --startup-project {path} {no_build} -- --environment {env}",
        "dotnet_ef_database_reset":
            "dotnet ef database update 0 --startup-project {path} {no_build} -- --environment {env}",
        "dotnet_ef_migrations_list":
            "dotnet ef migrations list --startup-project {path} --json {no_build} -- --environment {env}",
        "dotnet_ef_migrations_script":
            "dotnet ef migrations script {migration_from} --startup-project {path} --output {script_path} {no_build} "
            "{idempotent} -- --environment {env}",
        "dotnet_restore": "dotnet restore {force} {debug} {path}",
    }



"""Defines constants for the module."""

devops_toolset_download_resource = "https://github.com/aheadlabs/devops-toolset/archive/master.zip"
devops_toolset_save_as = "devops-toolset-master.zip"
devops_toolset_folder = "devops-toolset"
project_xml_name = "project.xml"
project_xml_download_resource = "https://raw.githubusercontent.com/aheadlabs/devops-toolset/master/project.xml"
dev_branch_regex = r'(refs\/heads\/)?dev|development'
release_branch_regex = r'(refs\/heads\/)?release\/(.*)*'
main_branch_regex = r'(refs\/heads\/)?main|master'



"""Utilities and tools for the devops toolset repository"""

import devops_toolset.tools.constants as constants
import requests
from zipfile import ZipFile
from devops_toolset.tools.xmlparser import XMLParser
import logging
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.core.app import App
import os.path
import devops_toolset.tools.git as git_tools
import pathlib
import shutil
from devops_toolset.project_types.wordpress.Literals import Literals as WordpressLiterals

app: App = App()
literals = LiteralsCore([WordpressLiterals])


def get_devops_toolset(destination_path: str):
    """ Downloads and extracts devops-toolset from its public resource

    It performs a download into the desired folder, also this method will extract and clean the result by renaming the
    final folder and deleting the original downloaded .zip source.

    Args:
        destination_path: The destination path of the download / extraction

    """
    logging.info(literals.get("wp_devops_toolset_obtaining").format(
        resource=constants.devops_toolset_download_resource)
    )
    response = requests.get(constants.devops_toolset_download_resource, allow_redirects=True)
    devops_toolset_path_file = os.path.join(destination_path, constants.devops_toolset_save_as)
    with open(devops_toolset_path_file, 'ab') as devops_toolset:
        devops_toolset.write(response.content)
    with ZipFile(devops_toolset_path_file, 'r') as zip_object:
        zip_object.extractall(destination_path)
    zip_extension = pathlib.Path(constants.devops_toolset_save_as).suffixes[0]
    destination_file_without_zip_extension = constants.devops_toolset_save_as.replace(zip_extension, '')
    old_destination_folder = os.path.join(destination_path, destination_file_without_zip_extension)
    final_destination_folder = os.path.join(destination_path, constants.devops_toolset_folder)
    os.rename(old_destination_folder, final_destination_folder)
    logging.info(literals.get("wp_devops_toolset_obtained").format(
        path=final_destination_folder)
    )
    os.remove(devops_toolset_path_file)
    git_tools.purge_gitkeep(pathlib.Path(destination_path).as_posix())


def update_devops_toolset(toolset_path: str):
    """ Compares and updates devops-toolset if needed

    Args:
        toolset_path: The path that will be checked
    """
    is_latest_version = compare_devops_toolset_version(toolset_path)
    if not is_latest_version:
        if os.path.exists(toolset_path):
            logging.warning(literals.get("wp_devops_toolset_needs_update"))
            # Remove current version
            shutil.rmtree(toolset_path)
        else:
            logging.warning(literals.get("wp_devops_toolset_not_found").format(path=toolset_path))
        get_devops_toolset(pathlib.Path(toolset_path).parent)


def compare_devops_toolset_version(toolset_path: str) -> bool:
    """ Returns true if version is the latest (no update required), false otherwise

    Args
        toolset_path: The path that will be checked
    """
    project_xml_path = os.path.join(toolset_path, constants.project_xml_name)
    if not os.path.exists(project_xml_path):
        return False
    xml_parser = XMLParser()
    # Get current version of devops-toolset
    xml_parser.parse_from_path(project_xml_path)
    current_version = xml_parser.get_attribute_value("version")
    logging.debug(literals.get("wp_current_version").format(version=current_version))
    # Get latest version from GitHub
    response = requests.get(constants.project_xml_download_resource, allow_redirects=True)
    xml_parser.parse_from_content(response.content.decode("utf-8"))
    latest_version = xml_parser.get_attribute_value("version")
    logging.debug(literals.get("wp_latest_version").format(version=latest_version))
    return current_version == latest_version


if __name__ == "__main__":
    help(__name__)



"""Bootstraps everything

Args:
    --skip-i18n: If present it will skip loading gettext
"""

import importlib.util
import pathlib
from devops_toolset.core.settings import Settings
import devops_toolset.i18n.loader
import devops_toolset.core.log_setup


class App(object):
    """App object that contains core settings and functionalities."""

    # Initialize settings
    settings: Settings = Settings()

    def __init__(self, skip_i18n: bool = False):
        """
        Args:
            skip_i18n: If True it does not load the gettext engine
        """

        # Load gettext
        if not skip_i18n:
            devops_toolset.i18n.loader.setup(self.settings)

        # Configure logging
        devops_toolset.core.log_setup.configure(self.settings.log_config_file_path)

    def load_platform_specific(self, name: str):
        module_path = pathlib.Path.joinpath(self.settings.platform_specific_path, f"{name}.py")
        spec = importlib.util.spec_from_file_location(name, module_path)
        platform_specific = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(platform_specific)
        return platform_specific


if __name__ == "__main__":
    help(__name__)



"""Tools for base64 encoding and decoding"""

import base64


def encode(text: str):
    """Encodes a string to base64 format

    Args:
        text: UTF-8 string to be encoded in base64 format
    """

    text_bytes = text.encode("utf-8")
    base64_bytes = base64.b64encode(text_bytes)
    return base64_bytes.decode("utf-8")


def decode(base64_text):
    """Decodes a string from base64 format

    Args:
        base64_text: Base64 string to be decoded to UTF-8 format
    """

    base64_text_bytes = base64_text.encode("utf-8")
    text_bytes = base64.b64decode(base64_text_bytes)
    return text_bytes.decode("utf-8")


if __name__ == "__main__":
    help(__name__)



""" Script used on devops flows to parse a theme's style.css metadata """

import argparse
import devops_toolset.filesystem.parsers as parsers


def main(path: str, tokens):
    """ Loads the css file in path and parses theme metadata, creating environment variables
    Args:
        path: Path to style.css file inside the target theme
        tokens: List of tokens to be retrieved from metadata file
    """
    with open(path, "rb") as css_file_content:
        parsers.parse_theme_metadata(css_file_content.read(), tokens, True)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("path", action=tools.argument_validators.PathValidator)
    args, args_unknown = parser.parse_known_args()
    # First parameter should be the path, all next parameters will be the tokens to retrieve
    main(args.path, args_unknown)



"""This script kicks off a CI pipeline for .NET projects"""

import argparse
import devops_toolset.configure
import devops_toolset.tools.cli

from devops_toolset.core.app import App
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.project_types.dotnet.entity_framework import check_branch_suitableness_for_migrations
from devops_toolset.project_types.dotnet.Literals import Literals as DotnetLiterals
from devops_toolset.project_types.dotnet.utils import get_csproj_project_version
from devops_toolset.tools.git import get_current_branch_simplified
from devops_toolset.tools.http_protocol import get_public_ip_address

app: App = App()
literals = LiteralsCore([DotnetLiterals])


def main(devops_platform: [str, None],
         csproj_path: [str, None],
         current_branch: [str, None],
         migrations_suitable_branches: [list, None],
         skip_get_public_ip_address: bool):
    """Kick off the CI pipeline

    Args:
        devops_platform: Name of the DevOps platform to be set.
        csproj_path: Path to the .csproj file.
        current_branch: Name of the current branch
        migrations_suitable_branches: List of suitable branches for migrations.
        skip_get_public_ip_address: Skips getting public IP address if present.
            simplified if present.
    """

    # Set DevOps current platform
    if devops_platform is not None:
        devops_toolset.configure.main(devops_platform=devops_platform, language="en")

    # Get project version from .csproj file
    if csproj_path is not None:
        _ = devops_toolset.project_types.dotnet.utils.get_csproj_project_version(csproj_path=csproj_path)

    # Get current branch simplified
    if current_branch is not None:
        _ = devops_toolset.tools.git.get_current_branch_simplified(branch=current_branch)

        # Check if Git branch is suitable for migrations
        if migrations_suitable_branches is not None:
            _ = devops_toolset.project_types.dotnet.entity_framework.check_branch_suitableness_for_migrations(
                current_simplified_branch=_,
                suitable_branches=migrations_suitable_branches
            )

    # Get public IP address
    if not skip_get_public_ip_address:
        _ = devops_toolset.tools.http_protocol.get_public_ip_address()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("devops_platform", default=None)
    parser.add_argument("--csproj-path", default=None)
    parser.add_argument("--current-branch", default=None)
    parser.add_argument("--migrations-suitable-branches", default=None)
    parser.add_argument("--skip-get-public-ip-address", action="store_true", default=False)
    args, args_unknown = parser.parse_known_args()

    devops_toolset.tools.cli.print_title(literals.get("dotnet_ci_title_pipeline_kickoff"))
    main(args.devops_platform,
         args.csproj_path,
         args.current_branch,
         args.migrations_suitable_branches,
         args.skip_get_public_ip_address)



"""XML toolset to interact and parse xml files"""

import xml.etree.ElementTree as ET
import os.path


class XMLParser(object):
    """ Class to to interact to xml files """

    def __init__(self):
        self.xml_file = ""

    def parse_from_path(self, xmlpath: str):
        """ Parses an xml file from a path and assign to xml_file
        Args
            xmlpath: Path to retrieve the xml file from
        """
        if os.path.exists(xmlpath):
            self.xml_file = ET.parse(xmlpath)

    def parse_from_content(self, content: str):
        """ Parses an xml file from an xml string content
                Args
                    content: The xml content to parse from
                """
        self.xml_file = ET.fromstring(content)

    def get_attribute_value(self, attribute_name: str) -> str:
        """ Gets a string containing a value of the attribute requested inside the xml_file
                Args
                    attribute_name: Name of the attribute which value will be recovered
                """
        if self.xml_file:
            return self.xml_file.find(attribute_name).text



"""linux module commands"""

from devops_toolset.core.ValueDictsBase import ValueDictsBase
from devops_toolset.core.app import App

app: App = App()


class Commands(ValueDictsBase):
    """Commands for the linux module."""

    # Add your linux commands dictionaries here
    _commands = {
        "create_env_variable": "export {variable_name}={variable_value}",
        "deb_which": "which {package}",
        "deb_package_install": "sudo apt install -y {package} {version}",
        "edit_in_place": "sed -i 's/{search_for}/{replace_with}/g' {file_path}"
    }



"""Microsoft Entity Framework utilities"""

from devops_toolset.core.app import App
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.project_types.dotnet.Literals import Literals as DotnetLiterals
from devops_toolset.core.CommandsCore import CommandsCore
from devops_toolset.project_types.dotnet.commands import Commands as DotnetCommands

import devops_toolset.project_types.dotnet.utils as utils
import devops_toolset.tools.cli as cli
import json
import logging
import pathlib
import re

app: App = App()
platform_specific = app.load_platform_specific("environment")
literals = LiteralsCore([DotnetLiterals])
commands = CommandsCore([DotnetCommands])


def __generate_sql_script(startup_project_path: str, script_path: str, environment: str, migration_from: str = "0",
                          no_build: bool = False, idempotent: bool = True) -> str:
    """ Generates a SQL script to apply ad hoc migrations to DBMS and returns its path

    Args:
        startup_project_path: Path to the startup project.
        script_path: Path where the script will be generated at.
        environment: Name for the environment to generate the script for.
        migration_from: ID / name of the last migration applied, defaults to 0.
        no_build: Skips build if True.
        idempotent: If True it creates an idempotent script.
    """

    logging.info(literals.get("dotnet_ef_migrations_script"))

    cli.call_subprocess(commands.get("dotnet_ef_migrations_script").format(
        migration_from=migration_from,
        path=startup_project_path,
        script_path=script_path,
        no_build="--no-build" if no_build else "",
        idempotent="--idempotent" if idempotent else "",
        env=environment
    ))

    return script_path


def __get_first_migration_not_applied(migrations_list: list) -> ([str, None], [str, None]):
    """ Returns the first migration in the list that has not been applied.

    Args:
        migrations_list: List of migrations

    Returns:
        Tuple with migration name and migration date (migration format) or None
        and None if no migration is found.
    """
    regex_pattern = r"^(\d+)_[\w-]+$"

    for migration in migrations_list:
        if migration["applied"] is False:
            match = re.search(regex_pattern, migration["id"])
            logging.info(literals.get("dotnet_ef_first_migration_not_applied").format(
                migration_name=migration["name"],
                migration_date=match.groups()[0]
            ))
            return migration["name"], match.groups()[0]

    logging.info(literals.get("dotnet_ef_no_pending_migrations"))
    return None, None


def __get_migrations_list(startup_project_path: str, environment: str, no_build: bool = False) -> list:
    """ Gets a list of the migrations applied for a specific environment

    Args:
        startup_project_path: Path to the startup project
        environment: Name for the environment to get the migrations for
        no_build: Skips build if True

    Returns:
        Migrations JSON array
    """

    logging.info(literals.get("dotnet_ef_migrations_list"))

    ef_command: str = commands.get("dotnet_ef_migrations_list").format(
        path=startup_project_path,
        no_build="--no-build" if no_build else "",
        env=environment
    )
    logging.info(literals.get("dotnet_ef_script_executing_command").format(command=ef_command))

    result: str = cli.call_subprocess_with_result(ef_command)
    logging.info(literals.get("dotnet_ef_migrations_list_output").format(output=result))

    migrations: str = ""
    json_lure: bool = False
    for line in result.splitlines():
        if line.startswith("[") and not json_lure:
            json_lure = True
        if json_lure:
            migrations += line

    return json.loads(migrations)


def __parse_data_from_migrations_json_array(migrations_json_array: list) -> (int, int, str):
    """ Parses data from the migrations JSON array

    Arguments:
        migrations_json_array: Migrations JSON array

    Returns:
        Number of migrations, number of applied migrations, last applied
        migration name
    """

    last_applied_migration: str = "0"
    applied_migrations: int = 0
    for migration in migrations_json_array:
        if migration["applied"]:
            last_applied_migration = migration["id"]
            applied_migrations += 1

    logging.info(literals.get("dotnet_ef_migrations_info").format(
        number=len(migrations_json_array),
        applied=applied_migrations,
        name=last_applied_migration
    ))
    return len(migrations_json_array), applied_migrations, last_applied_migration


def check_branch_suitableness_for_migrations(
        current_simplified_branch: str, suitable_branches: list,
        environment_variable_name: str = "DT_SUITABLE_BRANCH_FOR_MIGRATIONS") -> bool:
    """Checks if the current Git branch is suitable for generating and applying
    SQL migration scripts.

    Args:
        current_simplified_branch: Current simplified Git branch.
        suitable_branches: List of branches that are suitable for migrations.
        environment_variable_name: Name of the environment variable to be
            created. Defaults to "DT_SUITABLE_BRANCH_FOR_MIGRATIONS".

    Returns:
        True if branch is suitable, False otherwise.
    """

    is_suitable = True if current_simplified_branch in suitable_branches else False
    platform_specific.create_environment_variables({environment_variable_name: is_suitable})
    return is_suitable


def drop_database(startup_project_path: str, environment: str, no_build: bool = False):
    """ Drops the database

    Args:
        startup_project_path: Path to the startup project
        environment: Name for the environment to get the migrations for
        no_build: Skips build if True
    """
    logging.info(literals.get("dotnet_ef_database_drop"))

    result: str = cli.call_subprocess_with_result(commands.get("dotnet_ef_database_drop").format(
        path=startup_project_path,
        no_build="--no-build" if no_build else "",
        env=environment
    ))
    logging.info(result)


def generate_migration_sql_script(startup_project_path: str, environment: str, script_path: str,
                                  no_build: bool = False, idempotent: bool = True) -> str:
    """ Generates SQL migration script for a specific environment and returns its path

    Args:
        startup_project_path: Path to the startup project.
        environment: Name for the environment to get the migrations for.
        script_path: Path to the SQL script to be generated.
        no_build: Skips build if True.
        idempotent: If True it creates an idempotent script.
    """

    migrations_list: list = __get_migrations_list(startup_project_path, environment)
    migration_name, migration_date = __get_first_migration_not_applied(migrations_list)
    migrations, applied_migrations, last_migration_applied = __parse_data_from_migrations_json_array(migrations_list)

    if migration_name is not None and migrations != applied_migrations:
        return __generate_sql_script(startup_project_path, script_path.replace("#date#", migration_date), environment,
                                     last_migration_applied, no_build, idempotent)


def generate_migration_sql_scripts_for_all_environments(startup_project_path: str, scripts_base_path: str,
                                                        include_development: bool = False, no_build: bool = False,
                                                        idempotent: bool = True) -> list[str]:
    """ Generates a SQL migration script for every environment configured in
    the appsettings.*.json files and returns a list with the script paths generated

    Args:
        startup_project_path: Path to the startup project.
        scripts_base_path: Path to the directory where all scripts will be
            created at.
        include_development: If True, Development/Dev is included in the list.
        no_build: Skips build if True.
        idempotent: If True it creates an idempotent script.
    """
    script_paths = []
    environments = utils.get_appsettings_environments(startup_project_path, include_development)
    logging.info(literals.get("dotnet_ef_got_environments").format(environments=environments))

    base_path_obj = pathlib.Path(scripts_base_path)

    for environment in environments:
        logging.info(literals.get("dotnet_ef_script_for_environment").format(environment=environment))

        script_path = pathlib.Path.joinpath(base_path_obj, f"database-migration-{environment.lower()}-from-#date#.sql")
        logging.info(literals.get("dotnet_ef_script_being_generated").format(script_path=script_path))

        generated_script_path: str = \
            generate_migration_sql_script(startup_project_path, environment, str(script_path), no_build, idempotent)

        script_paths.append(generated_script_path)

    return script_paths


def reset_database(startup_project_path: str, environment: str, no_build: bool = False):
    """ Reverts all migrations applied, dropping all tables and truncating
    __efmigrationshistory database

    Args:
        startup_project_path: Path to the startup project.
        environment: Name for the environment to get the migrations for.
        no_build: Skips build if True.
    """
    logging.info(literals.get("dotnet_ef_database_reset"))

    result: str = cli.call_subprocess_with_result(commands.get("dotnet_ef_database_reset").format(
        path=startup_project_path,
        no_build="--no-build" if no_build else "",
        env=environment
    ))
    logging.info(result)


if __name__ == "__main__":
    help(__name__)



""" Contains wrappers for npm commands and tasks """

import devops_toolset.tools.cli as cli
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.core.CommandsCore import CommandsCore
from devops_toolset.project_types.node.commands import Commands as NodeCommands
from devops_toolset.project_types.node.Literals import Literals as NodeLiterals
from devops_toolset.core.app import App

app: App = App()
literals = LiteralsCore([NodeLiterals])
commands = CommandsCore([NodeCommands])


def convert_npm_extra_args(*args):
    """ Converts a list of variable args into extra args of a command -- <args> """
    if args[0]:
        return " -- " + " ".join(args)
    return ""


def convert_npm_parameter_if_present(value: bool):
    """ Converts a boolean value to a --if-present string."""
    if value:
        return "--if-present"
    return ""


def convert_npm_parameter_silent(value: bool):
    """ Converts a boolean value to a --silent string."""
    if value:
        return "--silent"
    return ""


def run_script(command: str, silent: bool = False, if_present: bool = False, *args):
    """ Wrapper for running arbitrary package scripts

    See Also:
        https://docs.npmjs.com/cli-commands/run-script.html
    Args:

        command: Command to run.
        silent: Prevents showing npm ERR! output on error.
        if_present: Prevents exiting with a non-zero exit code when the script is undefined.
        *args: Extra parameters will be passed as arguments to the subyacent npm run command.
    """
    cli.call_subprocess(commands.get("npm_run").format(
        command=command,
        silent=convert_npm_parameter_silent(silent),
        if_present=convert_npm_parameter_if_present(if_present),
        extra_args=convert_npm_extra_args(args)
    ), log_before_out=[literals.get("npm_run_before").format(task=command)],
     log_after_out=[literals.get("npm_run_after").format(task=command)],
     log_after_err=[literals.get("npm_run_error").format(task=command)])


def install(folder: str = ""):
    """ Installs a package, and any packages that it depends on

    See Also:
        https://docs.npmjs.com/cli/install

    Args:
        folder: Install the package in the directory as a symlink in the current project

    """
    cli.call_subprocess(commands.get("npm_install").format(
        folder=folder,
    ), log_before_out=[literals.get("npm_install_before")],
     log_after_out=[literals.get("npm_install_after")],
     log_after_err=[literals.get("npm_install_error")])


if __name__ == "__main__":
    help(__name__)



"""Constants for the devops_platforms module"""


class Urls(object):
    """URL constants"""
    GITHUB_RAW_CONTENT = "https://raw.githubusercontent.com/aheadlabs/devops-toolset/master/src/"
    DEVOPS_TOOLSET_MASTER = GITHUB_RAW_CONTENT + \
                            "devops_toolset/project_types/wordpress/default-files/default-project.xml"
    SONAR_QUALITY_GATE_PARTIAL_URL = "/api/qualitygates/project_status?projectKey="
    DEFAULT_PROJECT_XML = GITHUB_RAW_CONTENT + \
                          "devops_toolset/project_types/wordpress/default-files/default-project.xml"
    DEFAULT_README = GITHUB_RAW_CONTENT + "devops_toolset/project_types/wordpress/default-files/default-README.md"
    DEFAULT_GITIGNORE = GITHUB_RAW_CONTENT + "devops_toolset/project_types/wordpress/default-files/default.gitignore"
    DEFAULT_SITE_CONFIG = GITHUB_RAW_CONTENT + \
                          "devops_toolset/project_types/wordpress/default-files/default-localhost-site.json"
    DEFAULT_SITE_ENVIRONMENTS = GITHUB_RAW_CONTENT + \
                                "devops_toolset/project_types/wordpress/default-files/default-site-environments.json"
    DEFAULT_WORDPRESS_PROJECT_STRUCTURE = \
        GITHUB_RAW_CONTENT + \
        "devops_toolset/project_types/wordpress/default-files/default-wordpress-project-structure.json"
    DEFAULT_WORDPRESS_DEVELOPMENT_THEME_STRUCTURE = \
        GITHUB_RAW_CONTENT + \
        "devops_toolset/project_types/wordpress/default-files/default-wordpress-development-theme-structure.json"

    bootstrap_required_files = {
        "*site.json": DEFAULT_SITE_CONFIG,
        "*site-environments.json": DEFAULT_SITE_ENVIRONMENTS,
        "*project-structure.json": DEFAULT_WORDPRESS_PROJECT_STRUCTURE
    }



"""tools module commands"""

from devops_toolset.core.ValueDictsBase import ValueDictsBase
from devops_toolset.core.app import App

app: App = App()


class Commands(ValueDictsBase):
    """Commands for the tools module."""

    # Add your core literal dictionaries here
    _commands = {
        "git_init": "git init {path}",
        "git_add": "git add .",
        "git_commit_m": "git commit -m \"{message}\"",
        "git_push_tag": "git {auth} push origin {tag_name}",
        "git_push_tag_delete": "git {auth} push --delete origin {tag_name}",
        "git_tag_add": "git tag -a {tag_name} {commit_name} -m {tag_name}",
        "git_tag_delete": "git tag -d {tag_name}",
        "public_ip_address_service_url": "http://checkip.dyndns.org",
        "git_auth": "-c http.extraheader=\"AUTHORIZATION: {auth_header}\""
    }



"""Angular module literals"""

from devops_toolset.core.app import App
from devops_toolset.core.ValueDictsBase import ValueDictsBase

app: App = App()


class Literals(ValueDictsBase):
    """ValueDicts for the Angular module."""

    _info = {
        "angular_project_version": _("The project version is {version}"),
    }
    _errors = {}



"""Tools for working with dict objects"""

import re


def filter_keys(key_list: dict, regex: str) -> list:
    """Filters keys in a dictionary

    Args:
        key_list: dict object that contains the keys to be filtered.
        regex: Regular expression used to filter the keys.

    Returns:
        list with the filtered keys.
    """

    filtered_keys = filter(lambda x: re.search(regex, x) is not None, key_list)
    return list(filtered_keys)


if __name__ == "__main__":
    help(__name__)



"""Azure module literals"""

from devops_toolset.core.app import App
from devops_toolset.core.ValueDictsBase import ValueDictsBase

app: App = App()


class Literals(ValueDictsBase):
    """ValueDicts for the Azure module."""

    _info = {
        "azure_cli_command_output": _("I got this output from the Azure CLI command:\n{output}"),
        "azure_cli_executing_command": _("Executing command => {command}"),
        "azure_cli_logging_in_service_principal":
            _("Logging into Azure using service principal {service_principal} on tenant {tenant}"),
        "azure_cli_logging_out":
            _("Logging out from Azure (current logged in account)"),
    }
    _errors = {
        "azure_cli_db_mysql_flexible_server_execute_file_query_parameters_error":
            _("You must either pass a SQL file path or SQL query text to be executed.")
    }



"""Contains Apache utilities"""

import pathlib
from devops_toolset.core.app import App
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.project_types.linux.Literals import Literals as LinuxLiterals
from devops_toolset.core.CommandsCore import CommandsCore
from devops_toolset.project_types.linux.commands import Commands as LinuxCommands

app: App = App()
literals = LiteralsCore([LinuxLiterals])
commands = CommandsCore([LinuxCommands])


def generate_htaccess_file_based_basic_auth_file_for_user(realm_name: str, passwords_file_path: str, user_name: str,
                                                         htaccess_path: str):
    """Generates an .htacess file that claims for file-based basic
    authentication for a specific user.

    More info at: https://httpd.apache.org/docs/2.4/howto/auth.html

    Args:
        realm_name: Name for the authentication realm.
        passwords_file_path: Path to the password's file.
        user_name: user that will have access after authenticating.
        htaccess_path: Path to the directory where the .htaccess file will be
            generated.
    """

    htaccess_file_path: str = pathlib.Path.joinpath(pathlib.Path(htaccess_path), ".htaccess")
    lines: list = [
        "AuthType Basic\n",
        f"AuthName {realm_name}\n",
        "AuthBasicProvider file\n",
        f"AuthUserFile {passwords_file_path}\n",
        f"Require user {user_name}\n"
    ]

    with open(htaccess_file_path, "w") as file:
        file.writelines(lines)


if __name__ == "__main__":
    help(__name__)



"""Azure DevOps module commands"""

from devops_toolset.core.ValueDictsBase import ValueDictsBase
from devops_toolset.core.app import App

app: App = App()


class Commands(ValueDictsBase):
    """Commands for the Azure DevOps platform module."""

    # Add your Azure DevOps commands here
    _commands = {
        "azdevops_rest_get_build_list": "https://dev.azure.com/{organization}/{project}/_apis/build/builds",
        "azdevops_rest_get_build": "https://dev.azure.com/{organization}/{project}/_apis/build/builds/{build_id}/"
                                   "artifacts?artifactName={artifact_name}",
        "azdevops_cli_login": "{token} | az devops login --organization {organization}",
        "azdevops_cli_universal_download": "az artifacts universal download --feed {feed} --name {name} --path {path} "
                                       "--version {version} --organization {organization}",
    }



"""Literals for the package."""

from devops_toolset.core.ValueDictsBase import ValueDictsBase
from devops_toolset.core.app import App

app: App = App()


class LiteralsCore(ValueDictsBase):
    """Core literals for the package.


    1. Create an instance of the LiteralsCore class (this class).
    2. Pass as a parameter list all Literals classes you want to include.

        from devops_toolset.core.LiteralsCore import LiteralsCore
        from wordpress.Literals import Literals as WpLiterals

        literals = LiteralsCore([WpLiterals])

    3. Get a literal by it's key.

        literals.get("err_1")
    """

    # Add your core literal dictionaries here
    _debug = {
        "function_params": _("Parameters passed to the function:"),
    }
    _info = {}
    _errors = {}



"""Project setup"""
import pathlib
import setuptools
import devops_toolset.filesystem.parsers as parsers

root_path: pathlib.Path = pathlib.Path(__file__).parent

with open(pathlib.Path(root_path, "README.md"), "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open(pathlib.Path(root_path, "requirements.txt"), "r", encoding="utf-8") as req_file:
    install_requires = req_file.read().splitlines()

project_xml_parsed = parsers.parse_project_xml_data(False)
name = project_xml_parsed["PROJECT_NAME"]
version = project_xml_parsed["PROJECT_VERSION"]

setuptools.setup(
    name=name,
    version=version,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=install_requires,
    include_package_data=True,
    package_data={"devops_toolset.core": ["*.json"],
                  "devops_toolset.locales": ["**/LC_MESSAGES/*.mo"]},
    url='https://github.com/aheadlabs/devops-toolset/',
    license='https://github.com/aheadlabs/devops-toolset/blob/master/LICENSE',
    author='Ivan Sainz | Alberto Carbonell',
    author_email='aheadlabs@gmail.com',
    description='General purpose DevOps-related scripts and tools.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.9"
)



"""Parses files for getting information from them"""

import json
import logging
import pathlib
from devops_toolset.core.app import App
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.filesystem.Literals import Literals as FileSystemLiterals

app: App = App()
platform_specific = app.load_platform_specific("environment")
literals = LiteralsCore([FileSystemLiterals])


def parse_composer_json_data(add_environment_variables: bool = True, composer_json_path: str = None) -> dict:
    """Reads the composer.json file and returns a dict with its data.

    Properties/objects elements are upper cased, underscored and prepended with
        parent name.

    Args:
        add_environment_variables: If True it adds every element of the dict as
            an environment variable.
        composer_json_path: Path to the composer.json file.
    """

    if composer_json_path is None:
        composer_json_path = pathlib.Path.joinpath(app.settings.root_path, "composer.json")
    else:
        composer_json_path = pathlib.Path(composer_json_path)
    logging.info(literals.get("fs_composer_path_is").format(path=composer_json_path))
    with open(composer_json_path, "r") as composer_file:
        data = json.load(composer_file)

    environment_variables = {
        "PROJECT_NAME": data["name"],
        "PROJECT_VERSION": data["version"]
    }

    if add_environment_variables:
        platform_specific.create_environment_variables(environment_variables)

    return environment_variables


if __name__ == "__main__":
    help(__name__)



"""Contains wrappers for WP CLI commands"""

import datetime
import logging
import devops_toolset.tools.cli as cli
from devops_toolset.core.app import App
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.project_types.wordpress.Literals import Literals as WordpressLiterals
from devops_toolset.core.CommandsCore import CommandsCore
from devops_toolset.project_types.wordpress.commands import Commands as WordpressCommands
from enum import Enum

app: App = App()
literals = LiteralsCore([WordpressLiterals])
commands = CommandsCore([WordpressCommands])


class ValueType(Enum):
    """Defines value types for values at the wp-config.php file"""

    CONSTANT = 1,
    VARIABLE = 2


def add_update_option(option: dict, wordpress_path: str, debug: bool = False, update_permalinks: bool = True):
    """Adds a WordPress option or updates it if exists.

    Args:
        option: WordPress option.
        wordpress_path: Path to the WordPress installation.
        update_permalinks: If True updates the permalink structure.
        debug: It True logs debug information.
    """

    # Check if the option exists
    option_exists, existing_option_value = check_if_option_exists(option["name"], wordpress_path, debug)

    if option_exists:
        update_database_option(option["name"], option["value"], wordpress_path, debug, option["autoload"])
    else:
        add_database_option(option["name"], option["value"], wordpress_path, debug, option["autoload"])

    if option["name"] == "permalink_structure" and update_permalinks:
        cli.call_subprocess(commands.get("wpcli_rewrite_structure").format(
            structure=option["value"],
            path=wordpress_path,
            debug_info=convert_wp_parameter_debug(debug)
        ))


def add_database_option(option_name: str, option_value: str, wordpress_path: str,
                        debug_info: bool, autoload: bool = False):
    """Adds an option at the wp_options (*) table in the WordPress
    database using WP-CLI.

    (*) wp_ prefix could change, but this function will work anyway because
        its value is obtained from the wp-config.php configuration file.
    For more information see:
        https://developer.wordpress.org/cli/commands/option/update/

    Args:
        option_name: Name for the option.
        option_value: Value for the option.
        wordpress_path: Path to WordPress files.
        debug_info: Toggles debug info on the command.
        autoload: If True converts the value to yes, no otherwise.
    """

    # Check if the option exists
    option_exists, existing_option_value = check_if_option_exists(option_name, wordpress_path, debug_info)

    # Adds the option if it does not exist
    if not option_exists and check_if_option_is_valid(option_name, option_value, autoload):
        cli.call_subprocess(commands.get("wpcli_option_add").format(
            option_name=option_name,
            option_value=option_value,
            autoload=convert_wp_parameter_autoload(autoload),
            path=wordpress_path,
            debug_info=convert_wp_parameter_debug(debug_info)),
            log_before_process=[literals.get("wp_wpcli_option_add_before").format(
                option_name=option_name,
                option_value=option_value)],
            log_after_err=[literals.get("wp_wpcli_option_add_error").format(
                option_name=option_name,
                option_value=option_value)])


def check_if_option_exists(option_name: str, wordpress_path: str, debug_info: bool = False) -> (bool, str):
    """Checks if an option exists in the wp_options (*) table in the
    WordPress database using WP-CLI.

    (*) wp_ prefix could change, but this function will work anyway because
        its value is obtained from the wp-config.php configuration file.
    For more information see:
        https://developer.wordpress.org/cli/commands/option/get/

    Args:
        option_name: Name for the option.
        wordpress_path: Path to WordPress files.
        debug_info: Toggles debug info on the command.

    Returns:
        The option value if it exists
    """

    value = cli.call_subprocess_with_result(commands.get("wpcli_option_get").format(
        option_name=option_name,
        path=wordpress_path,
        debug_info=convert_wp_parameter_debug(debug_info)
    ))

    option_exists: bool = True if value is not None else False
    option_value: str = value.rstrip("\n") if option_exists else None

    return option_exists, option_value


def check_if_option_is_valid(name: str, value: str, autoload: bool) -> bool:
    """Checks if the name, value and autoload flag are valid (not empty
    or None)

    Args:
        name: Option name
        value: Option value
        autoload: Option autoload flag

    Returns:
        True if the option is valid, False otherwise
    """
    if name is not None and name != "" and value is not None \
            and autoload is not None:
        return True

    return False


def convert_wp_parameter_autoload(autoload: bool):
    """Converts a boolean value to a --autoload parameter."""
    if autoload:
        return "--autoload=yes"
    else:
        return "--autoload=no"


def convert_wp_parameter_db_user(db_user: str):
    """Converts a str value to a --db_user parameter."""
    if db_user:
        return "--dbuser=" + "\"" + db_user + "\""
    else:
        return ""


def convert_wp_parameter_db_pass(db_pass: str):
    """Converts a str value to a --db_pass parameter."""
    if db_pass:
        return "--dbpass=" + "\"" + db_pass + "\""
    else:
        return ""


def convert_wp_parameter_activate(activate: bool):
    """Converts a str value to a --activate."""
    if activate:
        return "--activate"
    else:
        return ""


def convert_wp_parameter_admin_password(admin_password: str):
    """Converts a str value to a --admin_password parameter."""
    if admin_password:
        return "--admin_password=" + "\"" + admin_password + "\""
    else:
        return ""


def convert_wp_parameter_content(value: bool):
    """Converts a boolean value to a yes/no string."""
    if not value:
        return "yes"
    return "no"


def convert_wp_parameter_debug(value: bool):
    """Converts a boolean value to a --debug string."""
    if value:
        return "--debug"
    return ""


def convert_wp_parameter_force(value: bool):
    """Converts a boolean value to a --force string."""
    if value:
        return "--force"
    return ""


def convert_wp_parameter_raw(value: bool):
    """Converts a boolean value to a --raw string."""
    if value:
        return "--raw"
    return ""


def convert_wp_parameter_send_email(value: bool):
    """Converts a boolean value to a --send-email string."""
    if value:
        return "--send-email"
    return ""


def convert_wp_parameter_skip_check(value: bool):
    """Converts a boolean value to a --skip-check string."""
    if value:
        return "--skip-check"
    return ""


def convert_wp_parameter_skip_content(value: bool):
    """Converts a boolean value to a --skip-content string."""
    if value:
        return "--skip-content"
    return ""


def convert_wp_parameter_skip_email(value: bool):
    """Converts a boolean value to a --skip-email string."""
    if value:
        return "--skip-email"
    return ""


def convert_wp_parameter_str_key_value(key: str, value: str, quoted: bool = False) -> str:
    """Converts a string value to a --role string.

    Args:
        key: WordPress parameter name.
        value: Value for the WordPress parameter.
        quoted: If True quotes the value

    Returns:
        Parameter build as --key=value
    """
    quote = "\"" if quoted else ""

    if value:
        return "--" + key + "=" + quote + value + quote
    return ""


def convert_wp_parameter_yes(value: bool):
    """Converts a boolean value to a --yes string."""
    if value:
        return "--yes"
    return ""


def create_configuration_file(wordpress_path: str, db_host: str, db_name: str, db_user: str,
                              db_pass: str, db_prefix: str, db_charset: str, db_collate: str,
                              skip_check: bool, debug: bool):
    """Creates the wp-config-php WordPress configuration file using WP-CLI.

    For more information see:
        https://developer.wordpress.org/cli/commands/config/create/

    Args:
        wordpress_path: Path to WordPress files
        db_host: Database host
        db_name: Database server
        db_user: Database user
        db_pass: Database password
        db_prefix: Prefix of database tables
        db_charset: Database charset
        db_collate: Database collation
        skip_check: Skip check parameter --skip-check
        debug: If present, --debug will be added to the command showing all debug trace information.
    """
    tools.cli.call_subprocess(commands.get("wpcli_config_create").format(
        path=wordpress_path,
        db_host=db_host,
        db_name=db_name,
        db_user=db_user,
        db_pass=db_pass,
        db_prefix=db_prefix,
        db_charset=db_charset,
        db_collate=db_collate,
        skip_check=convert_wp_parameter_skip_check(skip_check),
        debug_info=convert_wp_parameter_debug(debug)
    ), log_before_process=[literals.get("wp_wpcli_creating_config")],
        log_after_out=[literals.get("wp_wpcli_config_created_ok")],
        log_after_err=[literals.get("wp_wpcli_config_create_err")]
    )


def create_database(wordpress_path: str, debug: bool, db_user: str, db_password: str, schema: str):
    """ Calls wp db create with the parameters
    Args
        wordpress_path: Path to WordPress files
        debug: If True, --debug will be added to the command showing all debug trace information.
        db_user: Database user
        db_pass: Database password
        schema: Schema name to be checked for existence
    """
    # Check if the database exists
    output = cli.call_subprocess_with_result(commands.get("wpcli_db_query_db_exists").format(
        schema=schema,
        admin_user=db_user,
        admin_password=db_password,
        path=wordpress_path
    ))
    database_exists = True if output is not None and output.split("\r\n")[1] == '1' else False

    # Create the database
    if not database_exists:
        tools.cli.call_subprocess(commands.get("wpcli_db_create").format(
            path=wordpress_path,
            db_user=convert_wp_parameter_db_user(db_user),
            db_pass=convert_wp_parameter_db_pass(db_password),
            debug_info=convert_wp_parameter_debug(debug)
        ), log_before_process=[literals.get("wp_wpcli_db_create_before")])

    # Warn the user we will not create the database
    else:
        logging.warning(literals.get("mysql_db_exists_skipping_creation").format(schema=schema))


def create_user(user: dict, wordpress_path: str, debug: bool):
    """Creates a WordPress user.

    Args:
        user: User to be created based on #/definitions/user at
            https://dev.aheadlabs.com/schemas/json/wordpress-site-schema.json
        wordpress_path: Path to WordPress files.
        debug: If present, --debug will be added to the command showing all debug trace information.
    """

    cli.call_subprocess(commands.get("wp_user_create").format(
        user_login=user["user_login"],
        user_email=user["user_email"],
        role=convert_wp_parameter_str_key_value("role", user["role"]),
        display_name=convert_wp_parameter_str_key_value("display_name", user["display_name"], True),
        first_name=convert_wp_parameter_str_key_value("first_name", user["first_name"], True),
        last_name=convert_wp_parameter_str_key_value("last_name", user["last_name"], True),
        send_email=convert_wp_parameter_send_email(user["send_email"]),
        path=wordpress_path,
        debug_info=convert_wp_parameter_debug(debug)
    ),
        log_before_process=[literals.get("wp_wpcli_user_creating").format(user=user["user_login"])],
        log_after_out=[literals.get("wp_wpcli_user_created").format(user=user["user_login"])],
        log_after_err=[literals.get("wp_wpcli_user_creating_err").format(user=user["user_login"])]
    )


def user_exists(user_login: str, wordpress_path: str, debug: bool) -> bool:
    """Creates a WordPress user.

    Args:
        user_login: User login to be checked.
        wordpress_path: Path to WordPress files.
        debug: If present, --debug will be added to the command showing all debug trace information.

    Returns:
        True if the user exists.
    """

    result = cli.call_subprocess_with_result(commands.get("wp_user_get").format(
        user_login=user_login,
        path=wordpress_path,
        debug_info=convert_wp_parameter_debug(debug)
    ))

    if result is None:
        return False
    return True


def create_wordpress_database_user(wordpress_path: str, admin_user: str, admin_password: str, user: str, password: str,
                                   schema: str, host: str = 'localhost',
                                   db_privileges: str = 'create, alter, select, insert, update, delete',
                                   global_privileges: str = 'lock tables, process'):
    """Creates a database user to be used by WordPress
        e.g.:
            wp db query
                "create user '<username>'@'localhost'
                identified by '<password>'"
            wp db query
                "grant create, alter, select, insert, update, delete
                on <schema>.* to '<username>'@'localhost'"

        Args:
            wordpress_path: Path to WordPress files.
            admin_user: Database user with privileges to create databases and
                other users.
            admin_password: Admin user password.
            user: Database user name.
            password: Database user password.
            schema: Existing database schema name.
            host: localhost or FQDN. (% and _ wildcards are permitted).
            db_privileges: comma-separated privileges to be granted on the database. More info at:
                https://dev.mysql.com/doc/refman/en/grant.html#grant-privileges
                e.g.: 'create, alter, select, insert, update, delete'
            global_privileges: comma-separated privileges to be granted globally. More info at:
                https://dev.mysql.com/doc/refman/en/grant.html#grant-privileges
                e.g.: 'process'
    """
    # Check if the user exists
    output = cli.call_subprocess_with_result(commands.get("wpcli_db_query_user_exists").format(
        user=user,
        host=host,
        admin_user=admin_user,
        admin_password=admin_password,
        path=wordpress_path
    ))
    user_exists_db = True if output is not None and len(output.split("\r\n")) > 1 and output.split("\r\n")[1] == '1' \
        else False

    # Create user
    if not user_exists_db:
        cli.call_subprocess(commands.get("wpcli_db_query_create_user").format(
            user=user,
            host=host,
            password=password,
            admin_user=admin_user,
            admin_password=admin_password,
            path=wordpress_path),
            log_before_out=[literals.get("wp_wpcli_db_query_user_creating").format(user=user, host=host)],
            log_after_err=[literals.get("wp_wpcli_db_query_user_creating_err").format(user=user, host=host)]
        )

        # Grant user privileges on the database
        cli.call_subprocess(commands.get("wpcli_db_query_grant").format(
            privileges=db_privileges,
            schema=schema,
            user=user,
            host=host,
            admin_user=admin_user,
            admin_password=admin_password,
            path=wordpress_path),
            log_before_out=[literals.get("wp_wpcli_db_query_user_granting").format(
                user=user, host=host, schema=schema, privileges=db_privileges)],
            log_after_err=[literals.get("wp_wpcli_db_query_user_granting_err").format(
                user=user, host=host, schema=schema)]
        )

        # Grant user global privileges
        cli.call_subprocess(commands.get("wpcli_db_query_grant").format(
            privileges=global_privileges,
            schema="*",
            user=user,
            host=host,
            admin_user=admin_user,
            admin_password=admin_password,
            path=wordpress_path),
            log_before_out=[literals.get("wp_wpcli_db_query_user_granting").format(
                user=user, host=host, schema=schema, privileges=db_privileges)],
            log_after_err=[literals.get("wp_wpcli_db_query_user_granting_err").format(
                user=user, host=host, schema=schema)]
        )

    # Warn the user we are not altering users and permissions
    else:
        logging.warning(literals.get("mysql_user_exists_grant_privileges_manually").format(
            user=user,
            host=host,
            schema=schema,
            db_privileges=db_privileges,
            global_privileges=global_privileges
        ))


def delete_post_type_content(wordpress_path: str, content_type: str, debug_info: bool = False):
    """ Calls db in order to delete content from a concrete post type
    Args:
        wordpress_path: Path to WordPress files.
        content_type: Type of the content to be deleted
        debug_info: If true, --debug will be added to the command showing all debug trace information.
    """

    id_list = get_post_type_ids(wordpress_path, content_type)

    cli.call_subprocess(commands.get("wpcli_post_delete_post_type").format(
        id_list=id_list,
        path=wordpress_path,
        debug_info=convert_wp_parameter_debug(debug_info)),
        log_before_out=[literals.get("wp_wpcli_post_delete_post_type_before").format(post_type=content_type)],
        log_before_err=[literals.get("wp_wpcli_post_delete_post_type_err").format(post_type=content_type)])


def get_post_type_ids(wordpress_path: str, post_type: str):
    """Gets the ids for all the posts that match an specific post type.

    Args:
        wordpress_path: Path to WordPress files.
        post_type: Post type name to filter by.
    """

    return cli.call_subprocess_with_result(commands.get("wpcli_post_list_ids").format(
        post_type=post_type,
        path=wordpress_path
    ))


def wordpress_is_downloaded(path: str) -> bool:
    """Checks if WordPress is downloaded at the specified path.

    Args:
        path: Path to be checked for WordPress files.

    Returns:
        True if WordPress files are present at the specified path.
    """

    version = cli.call_subprocess_with_result(commands.get("wpcli_core_version").format(path=path))

    if version:
        logging.warning(literals.get("wp_wpcli_core_version_already_downloaded")
                        .format(version=version.replace("\n", "")))
        return True
    else:
        return False


def download_wordpress(destination_path: str, version: str, locale: str, skip_content: bool, debug: bool):
    """ Downloads the latest version of the WordPress core files using WP-CLI.

    For more information see:
        https://developer.wordpress.org/cli/commands/core/download/

    Args:
        destination_path: Path where WP-CLI will be downloaded.
        version: Target version to be downloaded (latest by default)
        locale: Wordpress locale
        skip_content: --skip-content parameter
        debug: If present, --debug will be added to the command showing all debug trace information.
    """
    if not wordpress_is_downloaded(destination_path):
        cli.call_subprocess(commands.get("wpcli_core_download").format(
            version=version,
            locale=locale,
            path=destination_path,
            skip_content=convert_wp_parameter_skip_content(skip_content),
            debug_info=convert_wp_parameter_debug(debug)
        ), log_before_process=[
            literals.get("wp_wpcli_downloading_wordpress").format(version=version, locale=locale, content=skip_content),
            literals.get("wp_wpcli_downloading_path").format(path=destination_path),
            literals.get("wp_wpcli_downloading_content").format(content=skip_content)],
            log_after_out=[
                literals.get("wp_wpcli_downloading_wordpress_ok")],
            log_after_err=[
                literals.get("wp_wpcli_downloading_wordpress_err")]
        )


def eval_code(php_code: str, wordpress_path: str) -> str:
    """ Executes a piece of php code
     Args:
        php_code: Piece of php code to be evaluated
        wordpress_path: Path to WordPress files.
    """
    return cli.call_subprocess_with_result(commands.get("wpcli_eval").format(
        php_code=php_code,
        path=wordpress_path))


def export_database(wordpress_path: str, dump_file_path: str, debug: bool):
    """Exports a WordPress database to a dump file using WP-CLI.

    All parameters are obtained from a site configuration file.

    For more information see:
        https://developer.wordpress.org/cli/commands/db/export/

    Args:
        wordpress_path: Path to WordPress files.
        dump_file_path: Path to the destination dump file.
        debug: If present, --debug will be added to the command showing all debug trace information.
    """
    cli.call_subprocess(commands.get("wpcli_db_export").format(
        core_dump_path=dump_file_path,
        path=wordpress_path,
        debug_info=convert_wp_parameter_debug(debug)),
        log_before_out=[literals.get("wp_wpcli_db_export_before").format(path=dump_file_path)],
        log_after_err=[literals.get("wp_wpcli_db_export_error")])


def export_content_to_wxr(wordpress_path: str, destination_path: str, wrx_file_suffix: str = None):
    """Exports all WordPress content to a WXR XML file.

    e.g.:
        wp export --dir="<destination path>" --filename_format="<date>-content.xml"

    Args:
        wordpress_path: Path to WordPress files.
        destination_path: Path where the files will be generated.
        wrx_file_suffix: Suffix to be added to the generated XML file.
    """
    date = datetime.datetime.utcnow().strftime("%Y.%m.%d")
    suffix = "" if wrx_file_suffix is None else f"-{wrx_file_suffix}"

    cli.call_subprocess(commands.get("wpcli_export").format(
        path=wordpress_path,
        destination_path=destination_path,
        date=date,
        suffix=suffix),
        log_before_out=[literals.get("wp_wpcli_export").format(path=destination_path)],
        log_after_err=[literals.get("wp_wpcli_export_err").format(path=destination_path)])


def import_database(wordpress_path: str, dump_file_path: str, debug: bool):
    """Imports a WordPress database from a dump file using WP-CLI.

    All parameters are obtained from a site configuration file.

    Args:
        wordpress_path: Path to WordPress files.
        dump_file_path: Path to dump file to be imported.
        debug: If present, --debug will be added to the command showing all debug trace information.
    """
    cli.call_subprocess(commands.get("wpcli_db_import").format(
        file=dump_file_path, path=wordpress_path, debug_info=convert_wp_parameter_debug(debug)),
                        log_before_process=[literals.get("wp_wpcli_db_import_before"), dump_file_path],
                        log_after_err=[literals.get("wp_wpcli_db_import_error")])


def import_wxr_content(wordpress_path: str, wxr_path: str, authors: str, debug: bool):
    """ Imports wxr files as content for the current WordPress from the wxr_path

    All parameters are obtained from a site configuration file.
    For more info, see: https://developer.wordpress.org/cli/commands/import/

     Args:
        wordpress_path: Path to WordPress files.
        wxr_path: Path the wxr files will be imported from.
        authors: Value for the authors argument.
        debug: If present, --debug will be added to the command showing all debug trace information.

    """
    cli.call_subprocess(commands.get("wpcli_import").format(
        file=wxr_path, path=wordpress_path, authors=authors, debug_info=convert_wp_parameter_debug(debug)),
        log_before_process=[literals.get("wp_wpcli_import_before"), wxr_path],
        log_after_err=[literals.get("wp_wpcli_import_error")])


def install_theme(wordpress_path: str, source: str, activate: bool, debug: bool, theme_name: str):
    """Installs WordPress's theme files (and child themes also) using WP-CLI.

    All parameters are obtained from a site configuration file.

    For more information see:
        https://developer.wordpress.org/cli/commands/theme/install/

    Args:

       wordpress_path: Path to the wordpress installation.
       source: The source of the theme, can be a path to a zip file, an url or a slug.
       activate: --activate present will activate the theme after installing it.
       debug: --debug present will show all debug trace information.
       theme_name: Name of the theme to be installed (just used for log purposes)

    """
    cli.call_subprocess(commands.get("wpcli_theme_install").format(
        path=wordpress_path,
        source=source,
        activate=convert_wp_parameter_activate(activate),
        debug_info=convert_wp_parameter_debug(debug)),
        log_before_process=[literals.get("wp_wpcli_theme_install_before").format(theme_name=theme_name)],
        log_after_err=[literals.get("wp_wpcli_theme_install_error").format(theme_name=theme_name)])


def install_plugin(plugin_name: str, wordpress_path: str, activate: bool, force: bool, source: str, debug: bool):
    """ Uses WP-CLI command to install a plugin with 'wp plugin install <source>'.

           All parameters are obtained from a site configuration file.

           For more information see:
               https://developer.wordpress.org/cli/commands/theme/install/

           Args:
               plugin_name: Plugin name / slug
               wordpress_path: Path to the wordpress installation.
               activate: Activates the plugin.
               force: Forces install by removing previous version of the plugin.
               source: Source of the installation.
               debug: Adds optional --debug parameter in order to better track the command result.
           """
    cli.call_subprocess(commands.get("wpcli_plugin_install").format(
        path=wordpress_path,
        activate=convert_wp_parameter_activate(activate),
        force=convert_wp_parameter_force(force),
        source=source,
        debug_info=convert_wp_parameter_debug(debug)
    ),
        log_before_process=[literals.get("wp_wpcli_plugin_install_before").format(plugin_name=plugin_name)],
        log_after_err=[literals.get("wp_wpcli_plugin_install_error").format(plugin_name=plugin_name)])


def install_wordpress_core(wordpress_path: str, url: str, title: str, admin_user: str, admin_email: str,
                           admin_password: str, skip_email: bool, debug: bool):
    """Installs WordPress core files using WP-CLI.

        All parameters are obtained from a site configuration file.

        For more information see:
            https://developer.wordpress.org/cli/commands/core/install/

        Args:
            wordpress_path: Path to WordPress files.
            url: The url for this wordpress site
            title: The title of this wordpress site
            admin_user: Name of the admin user of this site
            admin_email: Email if the admin user of this site
            admin_password: Password for the WordPress administrator user
            skip_email: --skip-mail parameter will send an email to the address specified if present.
            debug: Adds optional --debug parameter in order to better track the command result.
        """
    cli.call_subprocess(commands.get("wpcli_core_install").format(
        path=wordpress_path,
        url=url,
        title=title,
        admin_user=admin_user,
        admin_email=admin_email,
        admin_password=convert_wp_parameter_admin_password(admin_password),
        skip_email=convert_wp_parameter_skip_email(skip_email),
        debug_info=convert_wp_parameter_debug(debug)),
        log_before_process=[literals.get("wp_wpcli_core_install_before")],
        log_after_err=[literals.get("wp_wpcli_core_install_error")]
    )


def reset_database(wordpress_path: str, quiet: bool, debug_info: bool):
    """Removes all WordPress core tables from the database using WP-CLI.

    For more information see:
        https://developer.wordpress.org/cli/commands/db/reset/

    Args:
        debug_info:
        wordpress_path: Path to WordPress files.
        quiet: If True, no questions are asked.
    """
    cli.call_subprocess(commands.get("wpcli_db_reset").format(
        path=wordpress_path,
        yes=convert_wp_parameter_yes(quiet),
        debug_info=convert_wp_parameter_debug(debug_info)),
        log_before_process=[literals.get("wp_wpcli_db_reset_before")],
        log_after_err=[literals.get("wp_wpcli_db_reset_error")])


def reset_transients(wordpress_path: str):
    """Removes all WordPress transients from database using WP-CLI

    Args:
        wordpress_path: Path to WordPress files.
    """

    cli.call_subprocess(commands.get("wpcli_db_delete_transient").format(path=wordpress_path),
                        log_before_out=[literals.get("wp_wpcli_delete_transients")],
                        log_after_err=[literals.get("wp_wpcli_delete_transients_err")])


def set_configuration_value(name: str, value: str, value_type: ValueType, wordpress_path: str, raw: bool, debug: bool):
    """Creates or updates a value (constant or variable) at the wp-config-php
    WordPress configuration file using WP-CLI.

    For more information see:
        https://developer.wordpress.org/cli/commands/config/set/

        Args:
            name: Name of the parameter
            value: Value of the parameter
            value_type: CONSTANT or VARIABLE
            wordpress_path: Path to WordPress files.
            raw: Toggles --raw as parameter that decides if value will placed as it gets, without quotes
            debug: Toggles --debug_info as a parameter inside the command.
    """
    cli.call_subprocess(commands.get("wpcli_config_set").format(
        name=name,
        value=value,
        raw=convert_wp_parameter_raw(raw),
        type=value_type,
        path=wordpress_path,
        debug_info=convert_wp_parameter_debug(debug)
    ),  log_after_err=[literals.get("wp_wpcli_config_set_value_err")]
    )


def update_database_option(option_name: str, option_value: str, wordpress_path: str,
                           debug_info: bool, autoload: bool = False):
    """Updates an option at the wp_options (*) table in the WordPress
    database using WP-CLI.

    (*) wp_ prefix could change, but this function will work anyway because
        its value is obtained from the wp-config.php configuration file.
    For more information see:
        https://developer.wordpress.org/cli/commands/option/update/

    Args:
        option_name: Name for the option.
        option_value: Value for the option.
        wordpress_path: Path to WordPress files.
        debug_info: Toggles debug info on the command.
        autoload: If True converts the value to yes, no otherwise.
    """

    # Check if the option exists
    option_exists, existing_option_value = check_if_option_exists(option_name, wordpress_path, debug_info)

    # Update the option if the new value is different than the existing one
    if option_exists and existing_option_value != option_value:
        cli.call_subprocess(commands.get("wpcli_option_update").format(
            option_name=option_name,
            option_value=option_value,
            autoload=convert_wp_parameter_autoload(autoload),
            path=wordpress_path,
            debug_info=convert_wp_parameter_debug(debug_info)),
            log_before_process=[literals.get("wp_wpcli_option_update_before").format(
                option_name=option_name,
                option_value=option_value)],
            log_after_err=[literals.get("wp_wpcli_option_update_error").format(
                option_name=option_name,
                option_value=option_value)])
    else:
        logging.warning(literals.get("wp_wpcli_option_skipping").format(option_name=option_name))


def wp_cli_info():
    """Executes wp info command and logs output based on the result. """
    cli.call_subprocess(commands.get("wpcli_info"),
                        log_before_out=[literals.get("wp_wpcli_install_ok"), literals.get("wp_wpcli_info")],
                        log_after_out=[literals.get("wp_wpcli_add_ev")])


if __name__ == "__main__":
    help(__name__)



"""Provides common tools for all Azure services"""

from devops_toolset.core.app import App
from devops_toolset.core.CommandsCore import CommandsCore
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.project_types.azure.commands import Commands as AzureCommands
from devops_toolset.project_types.azure.Literals import Literals as AzureLiterals
from devops_toolset.tools import cli

import json
import logging


app: App = App()
literals = LiteralsCore([AzureLiterals])
commands = CommandsCore([AzureCommands])


def get_installed_cli_extensions() -> list:
    """Returns a list of the Azure extensions installed.

    Returns:
        List of Azure extensions.
    """

    az_command: str = commands.get("azure_cli_extension_list")
    logging.info(literals.get("azure_cli_executing_command").format(command=az_command))

    result: str = cli.call_subprocess_with_result(az_command)
    logging.info(literals.get("azure_cli_command_output").format(output=result))

    return json.loads(result)


def is_cli_extension_installed(name: str):
    """Checks if the provided Azure CLI extension is installed.

    Args:
        name: Extension name to be checked.
    """

    installed_extensions: list = get_installed_cli_extensions()
    extension_names = map(lambda ext: ext["name"], installed_extensions)

    return name in extension_names


def login_service_principal(user: str, secret: str, tenant: str) -> [list, None]:
    """Log into Azure using az login command.

    Args:
        user: Username for the service principal.
        secret: Secret for the service principal.
        tenant: Azure tenant ID.

    Returns:
        JSON list from Azure login.
    """

    az_command: str = commands.get("azure_cli_login_service_principal").format(
        user=user,
        secret=secret,
        tenant=tenant
    )

    result: str = cli.call_subprocess_with_result(az_command)
    logging.info(literals.get("azure_cli_logging_in_service_principal").format(
        service_principal=user,
        tenant=tenant
    ))

    if result:
        json_result = json.loads(result)
        logging.info(json_result)
        return json_result
    else:
        return None


def logout():
    """Log out from Azure using az logout command."""

    az_command: str = commands.get("azure_cli_logout")
    logging.info(literals.get("azure_cli_executing_command").format(command=az_command))

    cli.call_subprocess(az_command)
    logging.info(literals.get("azure_cli_logging_out"))


if __name__ == "__main__":
    help(__name__)



"""Logging configuration"""

import logging as logger
import logging.handlers
from logging.config import dictConfig
from devops_toolset.core.ColorFormatter import ColorFormatter
import devops_toolset.core.settings as settings

import json


def configure(filepath):
    """Configures the Python logging using a dictionary from a json file and adding a default
    fallback configuration with the basics

    Args:
        filepath: Path of the file which contains the configuration
        (See https://docs.python.org/2/library/logging.config.html#logging-config-api)
    """

    try:
        configure_by_file(filepath)
        add_filter_to_console_handler(logger.WARNING)
        add_colored_formatter_to_console_handlers()
    except Exception as err:
        logger.error(f"Cannot configure logger: {format(err)}")
        configure_by_default(logger.INFO)


def configure_by_default(loglevel):
    """ Configures a root logger using a default configuration

    Args:
        logging.level used by default
    """
    log = logger.getLogger()
    log.setLevel(loglevel)
    log.addHandler(logger.StreamHandler())
    log.info("Default configuration loaded successfully.")


def configure_by_file(filepath):
    """Configures the Python logging using a dictionary from a json file

    Args:
        filepath: Path of the file which contains the configuration
        (See https://docs.python.org/2/library/logging.config.html#logging-config-api)

    Raises:
        FileNotFoundError: If filepath doesn't exist.
        OSError: If filepath cannot be opened.
        AttributeError: If json config file content is not well-formed
    """
    with open(filepath, "r") as config_file:
        config = json.load(config_file)
        dictConfig(config["logging"])


def add_filter_to_console_handler(loglevel):
    """Adds a filter to the console in order to drop messages above desired loglevel

    Args:
        loglevel: Logging.loglevel that indicates messages below this level will be logged on this handler

    """
    log = logger.getLogger()
    handler = log.handlers[0]
    handler.addFilter(lambda record: record.levelno <= loglevel)


def add_time_rotated_file_handler(backupcount=10, filepath=".", when='midnight'):
    """Adds a filter to the console in order to drop messages above desired loglevel

     Args:
        backupcount: Number of backup files on rotation (10 files or days by default)
        filepath: Path of the file used to log in.
        when: Defines when to rotate
        (see https://docs.python.org/3/library/logging.handlers.html#timedrotatingfilehandler)

    Raises: FileNotFoundError: When passed filepath doesn't exist
    """
    log = logger.getLogger()
    file_handler = logging.handlers.TimedRotatingFileHandler(filename=filepath, when=when, backupCount=backupcount)
    log.addHandler(file_handler)


def add_colored_formatter_to_console_handlers():
    """ Adds a custom colored formatter to the current console handler """
    log = logger.getLogger()
    for handler in log.handlers:
        color_formatter = ColorFormatter(handler.formatter._fmt)
        handler.setFormatter(color_formatter)


if __name__ == "__main__":
    configure(settings.log_config_file_path)



""" This script will generate and configure a Wordpress site based on the
required configuration files"""

import argparse
import json
import logging
import pathlib
import requests
import devops_toolset.core.log_tools
import devops_toolset.filesystem.paths as paths
import os
import devops_toolset.project_types.wordpress.constants as constants
import devops_toolset.project_types.wordpress.wp_theme_tools as theme_tools
import devops_toolset.project_types.wordpress.wptools
import shutil
import devops_toolset.tools.cli as cli
import devops_toolset.tools.argument_validators
import devops_toolset.tools.devops_toolset_utils
import devops_toolset.tools.git as git_tools
from clint.textui import prompt
from devops_toolset.core.LiteralsCore import LiteralsCore
from devops_toolset.core.app import App
from devops_toolset.devops_platforms.constants import Urls
from devops_toolset.project_types.wordpress.Literals import Literals as WordpressLiterals

app: App = App()
literals = LiteralsCore([WordpressLiterals])


def main(root_path: str, db_user_password: str, db_admin_password: str, wp_admin_password: str,
         environment: str, additional_environments: list, environments_db_user_passwords: dict,
         create_db: bool, skip_partial_dumps: bool, create_development_theme: bool, **kwargs_):
    """Generates a new Wordpress site based on the site configuration file

    Args:
        root_path: Path to the repository root.
        db_user_password: Password for the database user.
        db_admin_password: Password for the database admin user.
        wp_admin_password: Password for the WordPress admin user.
        environment: Name of the environment to be processed.
        additional_environments: Additional environments to create additional
            wp-config.php files.
        environments_db_user_passwords: Additional environment db
            user passwords.
        create_db: If True it creates the database and the user.
        skip_partial_dumps: If True skips partial database dumps
            (after installing WordPress, themes and plugins).
        create_development_theme: If True generates the file structure for a
            development theme
        kwargs_: Platform-specific arguments
    """

    # Get basic settings
    global_constants: dict = devops_toolset.project_types.wordpress.wptools.get_constants()
    database_files_path: str = global_constants["paths"]["database"]
    root_path_obj: pathlib.Path = pathlib.Path(root_path)

    # Look for *site.json files in the project path
    required_files_pattern_suffixes: list = list(map(lambda x: f"*{x[1]}", constants.required_files_suffixes.items()))
    required_files_not_present: list[str] = paths.files_exist_filtered(
        root_path, False, required_files_pattern_suffixes)

    # If there are missing required files, ask for using the default ones from GitHub if quiet flag is activated
    if len(required_files_not_present) > 0:
        devops_toolset.core.log_tools.log_indented_list(literals.get("wp_required_files_not_found_detail")
                                                        .format(path=root_path),
                                                        required_files_not_present,
                                                        devops_toolset.core.log_tools.LogLevel.warning)

        devops_toolset.core.log_tools.log_indented_list(
            literals.get("wp_default_files"),
            [Urls.DEFAULT_SITE_ENVIRONMENTS,
             Urls.DEFAULT_SITE_CONFIG],
            devops_toolset.core.log_tools.LogLevel.info)

        # Ask to use defaults
        use_defaults: bool = prompt.yn(literals.get("wp_use_default_files"))
        # If not using defaults, exit
        if not use_defaults:
            logging.critical(literals.get("wp_required_files_mandatory"))
            raise ValueError(literals.get("wp_required_files_not_found").format(path=root_path))

        # Download defaults from GitHub
        for file in required_files_not_present:
            url = Urls.bootstrap_required_files[file]
            file_name = paths.get_file_name_from_url(url)
            file_path = pathlib.Path.joinpath(root_path_obj, file_name)

            response: requests.Response = requests.get(url)
            with open(file_path, "wb") as fw:
                fw.write(response.content)

    # Determine required file paths
    required_file_paths: tuple = devops_toolset.project_types.wordpress.wptools.get_required_file_paths(
        root_path, required_files_pattern_suffixes)

    # Parse configuration data
    site_config: dict = devops_toolset.project_types.wordpress.wptools.get_site_configuration(required_file_paths[0])
    environment_config = devops_toolset.project_types.wordpress.wptools.get_environment(site_config, environment)

    # Get future paths (from the constants.json file)
    wordpress_path: str = devops_toolset.project_types.wordpress.wptools.get_wordpress_path_from_root_path(
        root_path,
        global_constants)
    wordpress_path_as_posix: str = pathlib.Path(wordpress_path).as_posix()
    themes_path: str = theme_tools.get_themes_path_from_root_path(root_path, global_constants)

    # Create project structure & prepare devops-toolset
    devops_toolset.project_types.wordpress.wptools.start_basic_project_structure(root_path)

    # Check for updates / download devops-toolset
    setup_devops_toolset(root_path)

    # Download WordPress core files
    devops_toolset.project_types.wordpress.wptools.download_wordpress(site_config, wordpress_path,
                                                                      environment_config["wp_cli_debug"])

    # Create development theme (if needed)
    if create_development_theme:
        theme_tools.create_development_theme(site_config["settings"]["themes"], root_path, global_constants)

    # Set development themes / plugins ready
    theme_tools.build_theme(site_config["settings"]["themes"], themes_path, root_path)

    # Configure WordPress site
    devops_toolset.project_types.wordpress.wptools.set_wordpress_config_from_configuration_file(site_config,
                                                                                                environment_config,
                                                                                                wordpress_path,
                                                                                                db_user_password)

    # Create database and users
    if create_db:
        devops_toolset.project_types.wordpress.wptools.setup_database(environment_config, wordpress_path,
                                                                      db_user_password,
                                                                      db_admin_password)

    # Install WordPress site
    devops_toolset.project_types.wordpress.wptools.install_wordpress_site(
        site_config, environment_config, global_constants, root_path, wp_admin_password, skip_partial_dumps)

    # Add / update WordPress options
    devops_toolset.project_types.wordpress.wptools.add_wp_options(
        site_config["settings"]["options"], wordpress_path, environment_config["wp_cli_debug"])

    # Install site theme
    theme_tools.install_themes_from_configuration_file(
        site_config, environment_config, global_constants, root_path, skip_partial_dumps, **kwargs_)

    # Install site plugins
    devops_toolset.project_types.wordpress.wptools.install_plugins_from_configuration_file(
        site_config, environment_config, global_constants, root_path, skip_partial_dumps)

    # Create additional users
    devops_toolset.project_types.wordpress.wptools.create_users(site_config["settings"]["users"], wordpress_path,
                                                                environment_config["wp_cli_debug"])

    # Import wxr content
    if not create_development_theme:
        devops_toolset.project_types.wordpress.wptools.import_content_from_configuration_file(
            site_config, environment_config, root_path, wordpress_path, global_constants)

    # Generate additional wp-config.php files
    generate_additional_wpconfig_files(site_config, site_config["environments"], additional_environments,
                                       environments_db_user_passwords, wordpress_path)

    # Delete sample configuration file
    delete_sample_wp_config_file(wordpress_path)

    # Backup database
    core_dump_path_converted = devops_toolset.project_types.wordpress.wptools.convert_wp_config_token(
        site_config["settings"]["dumps"]["core"], wordpress_path)
    database_core_dump_directory_path = pathlib.Path.joinpath(root_path_obj, database_files_path)
    database_core_dump_path = pathlib.Path.joinpath(database_core_dump_directory_path, core_dump_path_converted)
    devops_toolset.project_types.wordpress.wptools.export_database(
        environment_config, wordpress_path_as_posix, database_core_dump_path.as_posix())
    git_tools.purge_gitkeep(database_core_dump_directory_path.as_posix())

    # Move config files to devops directory
    paths.move_files(
        root_path,
        pathlib.Path.joinpath(root_path_obj, global_constants["paths"]["devops"]).as_posix(),
        "*.json",
        False
    )


def setup_devops_toolset(root_path: str):
    """ Checks if devops toolset is present and up to date. In case not, it will be downloaded
    Args:
        root_path: Project's root path
    """
    devops_path_constant = devops_toolset.project_types.wordpress.wptools.get_constants()["paths"]["devops"]
    devops_path = pathlib.Path.joinpath(pathlib.Path(root_path), devops_path_constant, "devops-toolset").as_posix()
    logging.info(literals.get("wp_checking_devops_toolset").format(path=devops_path))
    devops_toolset.tools.devops_toolset_utils.update_devops_toolset(devops_path)


def generate_additional_wpconfig_files(site_config: dict, environments: dict, additional_environments: list,
                                       environments_db_user_passwords: dict,
                                       wordpress_path: str):
    """Generates additional wp-config.php files for different environments.

    Args:
        site_config: Parsed site configuration file
        environments: All the available environment configurations.
        additional_environments: Additional environments to generate
            wp-config.php files for.
        environments_db_user_passwords: Additional environment db user
            passwords.
        wordpress_path: Path to the WordPress installation.
    """

    wordpress_path_obj = pathlib.Path(wordpress_path)
    wp_config_path = pathlib.Path.joinpath(wordpress_path_obj, "wp-config.php")
    wp_config_path_temp = pathlib.Path.joinpath(wordpress_path_obj, "wp-config-temp.php")

    # Rename original file
    if paths.is_valid_path(wp_config_path.as_posix(), True):
        shutil.move(wp_config_path, wp_config_path_temp)

    # Filter environments
    filtered_environments = list(filter(lambda environment_x: environment_x["name"] in additional_environments,
                                        environments))

    # Create additional configuration files for the filtered environments
    for environment in filtered_environments:
        devops_toolset.project_types.wordpress.wptools.set_wordpress_config_from_configuration_file(
            site_config,
            environment,
            wordpress_path,
            environments_db_user_passwords[environment["name"]])
        shutil.move(wp_config_path, pathlib.Path.joinpath(wordpress_path_obj, f"wp-config-{environment['name']}.php"))

    # Rename original file
    if paths.is_valid_path(str(wp_config_path_temp), True):
        shutil.move(wp_config_path_temp, wp_config_path)


def delete_sample_wp_config_file(wordpress_path: str):
    """Deletes the wp-config-sample.php file.

    Args:
        wordpress_path: Path to WordPress.
    """
    file_path = pathlib.Path.joinpath(pathlib.Path(wordpress_path), "wp-config-sample.php")
    if file_path.exists():
        os.remove(str(file_path))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("project-path", action=devops_toolset.tools.argument_validators.PathValidator)
    parser.add_argument("--db-user-password", required=True)
    parser.add_argument("--db-admin-password", required=True)
    parser.add_argument("--wp-admin-password", required=True)
    parser.add_argument("--environment", default="localhost")
    parser.add_argument("--additional-environments", default="")
    parser.add_argument("--additional-environment-db-user-passwords", default={})
    parser.add_argument("--create-db", action="store_true", default=False)
    parser.add_argument("--skip-partial-dumps", action="store_true", default=False)
    parser.add_argument("--create-development-theme", action="store_true", default=False)
    args, args_unknown = parser.parse_known_args()
    kwargs = {}
    for kwarg in args_unknown:
        splited = str(kwarg).split("=")
        kwargs[splited[0]] = splited[1]

    # Parse long attributes to avoid lines longer that 120 characters
    additional_environment_db_user_passwords = args.additional_environment_db_user_passwords.split(",") \
        if args.additional_environment_db_user_passwords != "" else []

    cli.print_title(literals.get("wp_title_generate_wordpress"))
    main(args.project_path, args.db_user_password, args.db_admin_password, args.wp_admin_password,
         args.environment,
         args.additional_environments.split(",") if args.additional_environments != "" else [],
         json.loads(args.additional_environment_db_user_passwords),
         args.create_db,
         args.skip_partial_dumps,
         args.create_development_theme,
         **kwargs)
